/*
  COPYRIGHT NOTICE
  Copyright (c) 20011  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.analyze;

import edu.ucla.loni.Conversions;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.ByteOrder;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.MemoryCacheImageInputStream;

/**
 * Decoder for the header of the ANALYZE 7.5 format.
 *
 * @version 7 April 2011
 */
public class AnalyzeHeaderDecoder
{
    /** Decoded header. */
    private AnalyzeHeader _header;

    /** True if the encoded ANALYZE Header is byte swapped; false otherwise. */
    private boolean _isByteSwapped;

    /**
     * Constructs an ANALYZE Header Decoder.
     *
     * @param imageStream Input stream to the header.
     *
     * @throws IOException If an I/O error occurs.
     */
    public AnalyzeHeaderDecoder(ImageInputStream imageStream) throws IOException
    {
	// Read the header size
	imageStream.mark();
	int sizeOfHdr = imageStream.readInt();
	imageStream.reset();

	// Determine if bytes should be swapped; find the header size
	_isByteSwapped = false;
	if (sizeOfHdr == 0x94000000) {
	    _isByteSwapped = true;
	    sizeOfHdr = 148;
	}
	if (sizeOfHdr == 0x5C010000) {
	    _isByteSwapped = true;
	    sizeOfHdr = 348;
	}

	// Read the encoded Analyze header into a byte array
	byte[] headerBytes = new byte[sizeOfHdr];
	imageStream.readFully(headerBytes);

	// Create a new Image Input Stream that reads from the bytes
	ByteArrayInputStream bStream =new ByteArrayInputStream(headerBytes);
	MemoryCacheImageInputStream mStream =
	    new MemoryCacheImageInputStream(bStream);

	// Account for byte swapping if needed
	if (_isByteSwapped) {mStream.setByteOrder(ByteOrder.LITTLE_ENDIAN);}

	// Read Header Key elements
	byte[] dataType = new byte[10];
	byte[] dbName = new byte[18];

	sizeOfHdr = mStream.readInt();
	mStream.readFully(dataType);
	mStream.readFully(dbName);
	int extents = mStream.readInt();
	short sessionError = mStream.readShort();
	byte regular = mStream.readByte();
	byte hKeyUn0 = mStream.readByte();

	// Construct the Header Key
	HeaderKey headerKey =
	    new HeaderKey(sizeOfHdr, dataType, dbName, extents,
			  sessionError, regular, hKeyUn0);

	// Read Image Dimension elements
	short[] dim = new short[8];
	byte[] voxUnits = new byte[4];
	byte[] calUnits = new byte[8];
	short[] shorts = new short[4];
	float[] pixDim = new float[8];
	float[] floats = new float[8];

	mStream.readFully(dim, 0, dim.length);
	mStream.readFully(voxUnits);
	mStream.readFully(calUnits);
	mStream.readFully(shorts, 0, shorts.length);
	mStream.readFully(pixDim, 0, pixDim.length);
	mStream.readFully(floats, 0, floats.length);
	int glMax = mStream.readInt();
	int glMin = mStream.readInt();

	// Construct the Image Dimension
	ImageDimension imageDimension =
	    new ImageDimension(dim[0], dim[1], dim[2], dim[3], dim[4],
			       dim[5], dim[6], dim[7], voxUnits, calUnits,
			       shorts[0], shorts[1], shorts[2], shorts[3],
			       pixDim[0], pixDim[1], pixDim[2], pixDim[3],
			       pixDim[4], pixDim[5], pixDim[6], pixDim[7],
			       floats[0], floats[1], floats[2], floats[3],
			       floats[4], floats[5], floats[6], floats[7],
			       glMax, glMin);

	// Read Data History elements if present
	DataHistory dataHistory = null;
	if (sizeOfHdr == 348) {
	    byte[] descrip = new byte[80];
	    byte[] auxFile = new byte[24];
	    byte[] originator = new byte[10];
	    byte[] generated = new byte[10];
	    byte[] scannum = new byte[10];
	    byte[] patientId = new byte[10];
	    byte[] expDate = new byte[10];
	    byte[] expTime = new byte[10];
	    byte[] histUn0 = new byte[3];
	    int[] ints = new int[8];

	    mStream.readFully(descrip);
	    mStream.readFully(auxFile);
	    byte orient = mStream.readByte();

	    // Try to convert the originator bytes into a hex string
	    mStream.readFully(originator);
	    try {
		String string = Conversions.toHexString(originator);
		originator = string.getBytes();
	    }

	    // If not possible, treat as normal
	    catch (Exception e) {}

	    mStream.readFully(generated);
	    mStream.readFully(scannum);
	    mStream.readFully(patientId);
	    mStream.readFully(expDate);
	    mStream.readFully(expTime);
	    mStream.readFully(histUn0);
	    mStream.readFully(ints, 0, ints.length);

	    // Construct the Data History
	    dataHistory =
		new DataHistory(descrip, auxFile, orient, originator,
				generated, scannum, patientId, expDate,
				expTime, histUn0,
				ints[0], ints[1], ints[2], ints[3], ints[4],
				ints[5], ints[6], ints[7]);
	}

	// Construct the Analyze Header
	_header = new AnalyzeHeader(headerKey, imageDimension, dataHistory);
    }

    /**
     * Gets the decoded header.
     *
     * @return Decoded header.
     */
    public AnalyzeHeader getHeader()
    {
	return _header;
    }

    /**
     * Determines if the header is byte swapped.
     *
     * @return True if the encoded ANALYZE Header is byte swapped; false o/w.
     */
    public boolean isByteSwapped()
    {
	return _isByteSwapped;
    }
}
