/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.analyze;

import edu.ucla.loni.Conversions;
import java.io.IOException;
import java.util.Iterator;
import javax.imageio.stream.ImageOutputStream;

/**
 * Encoder for the header of the ANALYZE 7.5 format.
 *
 * @version 7 April 2011
 */
public class AnalyzeHeaderEncoder
{
    /** Output stream for the header. */
    private ImageOutputStream _imageStream;

    /**
     * Constructs an ANALYZE Header Encoder.
     *
     * @param imageStream Output stream for the header.
     */
    public AnalyzeHeaderEncoder(ImageOutputStream imageStream) 
    {
	_imageStream = imageStream;
    }

    /**
     * Writes the header.
     *
     * @param header Header to encode.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void writeHeader(AnalyzeHeader header) throws IOException
    {
	// Get the Analyze Header elements
	AnalyzeSubheader headerKey = header.getHeaderKey();
	AnalyzeSubheader imageDimension = header.getImageDimension();
	AnalyzeSubheader dataHistory = header.getDataHistory();

	// Determine the size of the Analyze Header in bytes
	int headerSize = 348;
	if (dataHistory == null) { headerSize = 148; }

	// Write the Header Key to the metadata output stream
	_writeHeaderKey(headerKey, _imageStream, headerSize);

	// Write the Image Dimension to the metadata output stream
	_writeImageDimension(imageDimension, _imageStream);

	// If applicable, write the Data History to the metadata output stream
	if (dataHistory != null) {
	    _writeDataHistory(dataHistory, _imageStream);
	}
    }

    /**
     * Writes the Header Key to the output stream.
     *
     * @param headerKey Header Key to write to the output stream.
     * @param outputStream Image Output Stream to write to.
     * @param headerSize Size of the Analyze header in bytes.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeHeaderKey(AnalyzeSubheader headerKey,
				 ImageOutputStream outputStream, int headerSize)
	throws IOException
    {
	Iterator iter = headerKey.getElementNames();
	while ( iter.hasNext() ) {
	    String elementName = (String)iter.next();
	    Object elementValue = headerKey.getElementValue(elementName);

	    // Element value is a String
	    if (elementValue instanceof String) {
		int size = 0;
		if (HeaderKey.DATA_TYPE.equals(elementName) ) { size = 10; }
		if (HeaderKey.DB_NAME.equals(elementName) ) { size = 18; }

		// Write the string value
		_writeString( (String)elementValue, size, outputStream );
	    }

	    // Element value is a Number
	    else {

		// Overwrite any value for the size of the Analyze Header
		if ( HeaderKey.SIZEOF_HDR.equals(elementName) ) {
		    elementValue = new Integer(headerSize);
		}

		// Write the number value
		_writeNumber( (Number)elementValue, outputStream );
	    }
	}
    }

    /**
     * Writes the Image Dimension to the output stream.
     *
     * @param imageDimension Image Dimension to write to the output stream.
     * @param outputStream Image Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeImageDimension(AnalyzeSubheader imageDimension,
				      ImageOutputStream outputStream)
	throws IOException
    {
	Iterator iter = imageDimension.getElementNames();
	while ( iter.hasNext() ) {
	    String elementName = (String)iter.next();
	    Object elementValue = imageDimension.getElementValue(elementName);

	    // Element value is a String
	    if (elementValue instanceof String) {
		int size = 0;
		if (ImageDimension.VOX_UNITS.equals(elementName) ) { size = 4; }
		if (ImageDimension.CAL_UNITS.equals(elementName) ) { size = 8; }

		// Write the string value
		_writeString( (String)elementValue, size, outputStream );
	    }

	    // Element value is a Number
	    else { _writeNumber( (Number)elementValue, outputStream ); }
	}
    }

    /**
     * Writes the Data History to the output stream.
     *
     * @param dataHistory Data History to write to the output stream.
     * @param outputStream Image Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeDataHistory(AnalyzeSubheader dataHistory,
				   ImageOutputStream outputStream)
	throws IOException
    {
	Iterator iter = dataHistory.getElementNames();
	while ( iter.hasNext() ) {
	    String elementName = (String)iter.next();
	    Object elementValue = dataHistory.getElementValue(elementName);

	    // Element value is a String
	    if (elementValue instanceof String) {
		int size = 0;
		if (DataHistory.DESCRIP.equals(elementName) ) { size = 80; }
		if (DataHistory.AUX_FILE.equals(elementName) ) { size = 24; }
		if (DataHistory.ORIGINATOR.equals(elementName) ) {
		    size = 10;

		    // Try to convert from a hex string 
		    try {
			byte[] bytes =
			    Conversions.toByteArray( (String)elementValue );

			elementValue = new String(bytes);
		    }

		    // If not an encoded hex string, write as normal
		    catch (Exception e) {}
		}
		if (DataHistory.GENERATED.equals(elementName) ) { size = 10; }
		if (DataHistory.SCAN_NUM.equals(elementName) ) { size = 10; }
		if (DataHistory.PATIENT_ID.equals(elementName) ) { size = 10; }
		if (DataHistory.EXP_DATE.equals(elementName) ) { size = 10; }
		if (DataHistory.EXP_TIME.equals(elementName) ) { size = 10; }
		if (DataHistory.HIST_UN0.equals(elementName) ) { size = 3; }

		// Write the string value
		_writeString( (String)elementValue, size, outputStream );
	    }

	    // Element value is a Number
	    else { _writeNumber( (Number)elementValue, outputStream ); }
	}
    }

    /**
     * Writes the Number to the output stream.
     *
     * @param number Number to write to the output stream.
     * @param outputStream Image Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeNumber(Number number, ImageOutputStream outputStream)
	throws IOException
    {
	// Byte
	if (number instanceof Byte) {
	    int value = ((Byte)number).byteValue() & 0xffffffff;
	    outputStream.writeByte(value);
	}

	// Short
	else if (number instanceof Short) {
	    int value = ((Short)number).shortValue() & 0xffffffff;
	    outputStream.writeShort(value);
	}

	// Integer
	else if (number instanceof Integer) {
	    outputStream.writeInt( ((Integer)number).intValue() );
	}

	// Float
	else if (number instanceof Float) {
	    outputStream.writeFloat( ((Float)number).floatValue() );
	}

	// Unknown Number
	else {
	    String msg = "Unable to write an Number of type \"" +
		number.getClass().getName() + "\".";
	    throw new IOException(msg);
	}
    }

    /**
     * Writes the String to the output stream.
     *
     * @param string String to write to the output stream.
     * @param length Length of the string to actually write.
     * @param outputStream Image Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeString(String string, int length,
			      ImageOutputStream outputStream)
	throws IOException
    {
	// Create a byte buffer of the specified length
	byte[] buffer = new byte[length];

	// Copy the bytes of the String to the byte buffer
	byte[] stringBytes = string.getBytes();
	int min = Math.min( buffer.length, stringBytes.length );
	System.arraycopy(stringBytes, 0, buffer, 0, min);

	// Write the bytes
	outputStream.write(buffer);
    }
}
