/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.ecat;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import javax.imageio.stream.ImageInputStream;

/**
 * Decoder for the header of the ECAT 7 format.
 *
 * @version 8 April 2011
 */
public class Ecat7HeaderDecoder
{
    /** Decoded header. */
    private Ecat7Header _header;

    /** Offsets (in bytes) to the start of each frame. */
    private long[] _frameOffsets;

    /**
     * Constructs an ECAT 7 Header Decoder.
     *
     * @param imageStream Input stream to the header.
     *
     * @throws IOException If an I/O error occurs.
     */
    public Ecat7HeaderDecoder(ImageInputStream imageStream) throws IOException
    {
	// Read the first metadata block
	byte[] mainHeaderBytes = new byte[512];
	imageStream.readFully(mainHeaderBytes);

	// Decode the main header
	Ecat7MainHeader mainHeader = getEcat7MainHeader(mainHeaderBytes);
	_header = new Ecat7Header(mainHeader);

	// Determine the number of frames to decode
	int numberOfFrames = _header.getNumberOfFrames();
	_frameOffsets = new long[numberOfFrames];
	int[] matrixIds = new int[numberOfFrames];

	// Decode all the frames
	int frameNumber = 0;
	while (frameNumber < numberOfFrames) {

	    // Read in the directory list block
	    byte[] listBytes = new byte[512];
	    imageStream.readFully(listBytes);
	    ByteArrayInputStream bStream = new ByteArrayInputStream(listBytes);
	    DataInputStream listStream = new DataInputStream(bStream);

	    // Decode the first entry of the directory list block
	    // (4 bytes) - Number of available entries in this dir list block
	    // (4 bytes) - Offset to next directory list block
	    // (4 bytes) - Offset to previous directory list block
	    // (4 bytes) - Number of used entries in this directory list block
	    int nfree = listStream.readInt();
	    int nxtblk = listStream.readInt();
	    int prvblk = listStream.readInt();
	    int nDims = listStream.readInt();

	    // Decode all the frame offsets in the directory list block
	    int numberOfEntries = Math.min(31, numberOfFrames-frameNumber);
	    for (int i = 0; i < numberOfEntries; i++) {

		// Decode the entry of the directory list block
		// (4 bytes) - Matrix ID
		// (4 bytes) - Matrix Subheader block number
		// (4 bytes) - Block number of last block of the image data
		// (4 bytes) - Status: 1=exists, 0=doesn't exist, -1=deleted
		int matnum = listStream.readInt();
		int strtblk = listStream.readInt();
		int endblk = listStream.readInt();
		int matsat = listStream.readInt();

		// Record the frame offset and matrix id
		_frameOffsets[frameNumber] = 512*(strtblk-1);
		matrixIds[frameNumber] = matnum;
		frameNumber++;
	    }

	    // Advance to the next directory list block
	    if (frameNumber < numberOfFrames) {
		imageStream.seek(512*(nxtblk-1));
	    }
	}

	// Decode the frame subheaders
	for (int i = 0; i < _frameOffsets.length; i++) {

	    // Move the beginning of the frame
	    imageStream.seek(_frameOffsets[i]);

	    // Read the frame block
	    byte[] imgHeaderBytes = new byte[512];
	    imageStream.readFully(imgHeaderBytes);

	    // Decode the frame subheader
	    Ecat7ImageHeader imageHeader = getEcat7ImageHeader(matrixIds[i],
							       imgHeaderBytes);
	    _header.setSubheader(imageHeader, i);
	}
    }

    /**
     * Gets the decoded header.
     *
     * @return Decoded header.
     */
    public Ecat7Header getHeader()
    {
	return _header;
    }

    /**
     * Gets the frame offsets.
     *
     * @return Offsets (in bytes) to the start of each frame.
     */
    public long[] getFrameOffsets()
    {
	return _frameOffsets;
    }

    /**
     * Reads an ECAT 7 Main Header from the specified byte source.
     *
     * @param mainHeaderBytes Byte array containing the encoded main header.
     *
     * @return ECAT 7 Main Header decoded from the byte source.
     * 
     * @throws IOException If an error occurs during reading.
     */
    public static Ecat7MainHeader getEcat7MainHeader(byte[] mainHeaderBytes)
	throws IOException
    {
	ByteArrayInputStream bStream=new ByteArrayInputStream(mainHeaderBytes);
	DataInputStream headerStream = new DataInputStream(bStream);

	// Define the byte arrays
	byte[] magicNumber = new byte[14];
	byte[] originalFileName = new byte[32];
	byte[] serialNumber = new byte[10];
	byte[] isotopeName = new byte[8];
	byte[] radiopharmaceutical = new byte[32];
	byte[] studyType = new byte[12];
	byte[] patientId = new byte[16];
	byte[] patientName = new byte[32];
	byte[] patientSex = new byte[1];
	byte[] patientDexterity = new byte[1];
	byte[] physicianName = new byte[32];
	byte[] operatorName = new byte[32];
	byte[] studyDescription = new byte[32];
	byte[] facilityName = new byte[20];
	byte[] userProcessCode = new byte[10];
	byte[] dataUnits = new byte[32];

	// Define other arrays
	float[] bedPosition = new float[15];
	short[] fillCti = new short[6];

	// Read info from the stream
	headerStream.readFully(magicNumber);
	headerStream.readFully(originalFileName);
	short swVersion = headerStream.readShort();
	short systemType = headerStream.readShort();
	short fileType = headerStream.readShort();
	headerStream.readFully(serialNumber);
	int scanStartTime = headerStream.readInt();
	headerStream.readFully(isotopeName);
	float isotopeHalflife = headerStream.readFloat();
	headerStream.readFully(radiopharmaceutical);
	float gantryTilt = headerStream.readFloat();
	float gantryRotation = headerStream.readFloat();
	float bedElevation = headerStream.readFloat();
	float intrinsicTilt = headerStream.readFloat();
	short wobbleSpeed = headerStream.readShort();
	short transmSourceType = headerStream.readShort();
	float distanceScanned = headerStream.readFloat();
	float transaxialFov = headerStream.readFloat();
	short angularCompression = headerStream.readShort();
	short coinSampMode = headerStream.readShort();
	short axialSampMode = headerStream.readShort();
	float ecatCalibrationFactor = headerStream.readFloat();
	short calibrationUnits = headerStream.readShort();
	short calibrationUnitsLabel = headerStream.readShort();
	short compressionCode = headerStream.readShort();
	headerStream.readFully(studyType);
	headerStream.readFully(patientId);
	headerStream.readFully(patientName);
	headerStream.readFully(patientSex);
	headerStream.readFully(patientDexterity);
	float patientAge = headerStream.readFloat();
	float patientHeight = headerStream.readFloat();
	float patientWeight = headerStream.readFloat();
	int patientBirthDate = headerStream.readInt();
	headerStream.readFully(physicianName);
	headerStream.readFully(operatorName);
	headerStream.readFully(studyDescription);
	short acquisitionType = headerStream.readShort();
	short patientOrientation = headerStream.readShort();
	headerStream.readFully(facilityName);
	short numPlanes = headerStream.readShort();
	short numFrames = headerStream.readShort();
	short numGates = headerStream.readShort();
	short numBedPos = headerStream.readShort();
	float initBedPosition = headerStream.readFloat();

	for (int i = 0; i < bedPosition.length; i++) {
	    bedPosition[i] = headerStream.readFloat();
	}
      
	float planeSeparation = headerStream.readFloat();
	short lwrSctrThres = headerStream.readShort();
	short lwrTrueThres = headerStream.readShort();
	short uprTrueThres = headerStream.readShort();
	headerStream.readFully(userProcessCode);
	short acquisitionMode = headerStream.readShort();
	float binSize = headerStream.readFloat();
	float branchingFraction = headerStream.readFloat();
	int doseStartTime = headerStream.readInt();
	float dosage = headerStream.readFloat();
	float wellCounterCorrFactor = headerStream.readFloat();
	headerStream.readFully(dataUnits);
	short septaState = headerStream.readShort();

	for (int i = 0; i < fillCti.length; i++) {
	    fillCti[i] = headerStream.readShort();
	}

	// Return an ECAT 7 Main Header
	return new Ecat7MainHeader(magicNumber, originalFileName, 
				   swVersion, systemType, fileType,
				   serialNumber, scanStartTime,
				   isotopeName, isotopeHalflife,
				   radiopharmaceutical, gantryTilt,
				   gantryRotation, bedElevation,
				   intrinsicTilt, wobbleSpeed,
				   transmSourceType, distanceScanned,
				   transaxialFov, angularCompression,
				   coinSampMode, axialSampMode,
				   ecatCalibrationFactor, calibrationUnits,
				   calibrationUnitsLabel, compressionCode,
				   studyType, patientId,
				   patientName, patientSex,
				   patientDexterity, patientAge,
				   patientHeight, patientWeight,
				   patientBirthDate, physicianName,
				   operatorName, studyDescription,
				   acquisitionType, patientOrientation,
				   facilityName, numPlanes, numFrames,
				   numGates, numBedPos,
				   initBedPosition, bedPosition,
				   planeSeparation, lwrSctrThres,
				   lwrTrueThres, uprTrueThres,
				   userProcessCode, acquisitionMode,
				   binSize, branchingFraction,
				   doseStartTime, dosage,
				   wellCounterCorrFactor, dataUnits,
				   septaState, fillCti);
    }

    /**
     * Reads an ECAT 7 Image Header from the specified byte source.
     *
     * @param matrixId Matrix id.
     * @param imageHeaderBytes Byte array containing the encoded image header.
     *
     * @return ECAT 7 Image Header decoded from the byte source.
     * 
     * @throws IOException If an error occurs during reading.
     */
    public static Ecat7ImageHeader getEcat7ImageHeader(int matrixId,
						       byte[] imageHeaderBytes)
	throws IOException
    {
	ByteArrayInputStream bStrm = new ByteArrayInputStream(imageHeaderBytes);
	DataInputStream headerStream = new DataInputStream(bStrm);

	// Define the byte arrays
	byte[] annotation = new byte[40];
	byte[] fillCti = new byte[2*87];
	byte[] fillUser = new byte[2*49];

	// Read info from the stream
	short dataType = headerStream.readShort();
	short numDimensions = headerStream.readShort();
	short xDimension = headerStream.readShort();
	short yDimension = headerStream.readShort();
	short zDimension = headerStream.readShort();
	float xOffset = headerStream.readFloat();
	float yOffset = headerStream.readFloat();
	float zOffset = headerStream.readFloat();
	float reconZoom = headerStream.readFloat();
	float scaleFactor = headerStream.readFloat();
	short imageMin = headerStream.readShort();
	short imageMax = headerStream.readShort();
	float xPixelSize = headerStream.readFloat();
	float yPixelSize = headerStream.readFloat();
	float zPixelSize = headerStream.readFloat();
	int frameDuration = headerStream.readInt();
	int frameStartTime = headerStream.readInt();
	short filterCode = headerStream.readShort();
	float xResolution = headerStream.readFloat();
	float yResolution = headerStream.readFloat();
	float zResolution = headerStream.readFloat();
	float numRElements = headerStream.readFloat();
	float numAngles = headerStream.readFloat();
	float zRotationAngle = headerStream.readFloat();
	float decayCorrFctr = headerStream.readFloat();
	int processingCode = headerStream.readInt();
	int gateDuration = headerStream.readInt();
	int rWaveOffset = headerStream.readInt();
	int numAcceptedBeats = headerStream.readInt();
	float filterCutoffFrequency = headerStream.readFloat();
	float filterResolution = headerStream.readFloat();
	float filterRampSlope = headerStream.readFloat();
	short filterOrder = headerStream.readShort();
	float filterScatterFraction = headerStream.readFloat();
	float filterScatterSlope = headerStream.readFloat();
	headerStream.readFully(annotation);
	float mt11 = headerStream.readFloat();
	float mt12 = headerStream.readFloat();
	float mt13 = headerStream.readFloat();
	float mt21 = headerStream.readFloat();
	float mt22 = headerStream.readFloat();
	float mt23 = headerStream.readFloat();
	float mt31 = headerStream.readFloat();
	float mt32 = headerStream.readFloat();
	float mt33 = headerStream.readFloat();
	float rFilterCutoff = headerStream.readFloat();
	float rFilterResolution = headerStream.readFloat();
	short rFilterCode = headerStream.readShort();
	short rFilterOrder = headerStream.readShort();
	float zFilterCutoff = headerStream.readFloat();
	float zFilterResolution = headerStream.readFloat();
	short zFilterCode = headerStream.readShort();
	short zFilterOrder = headerStream.readShort();
	float mt14 = headerStream.readFloat();
	float mt24 = headerStream.readFloat();
	float mt34 = headerStream.readFloat();
	short scatterType = headerStream.readShort();
	short reconType = headerStream.readShort();
	short reconViews = headerStream.readShort();
	headerStream.readFully(fillCti);
	headerStream.readFully(fillUser);

	// Return an ECAT 7 Image Header
	return new Ecat7ImageHeader(matrixId, dataType, numDimensions,
				    xDimension, yDimension, zDimension,
				    xOffset, yOffset, zOffset,
				    reconZoom, scaleFactor, imageMin,
				    imageMax, xPixelSize, yPixelSize,
				    zPixelSize, frameDuration,
				    frameStartTime, filterCode,
				    xResolution, yResolution,
				    zResolution, numRElements,
				    numAngles, zRotationAngle,
				    decayCorrFctr, processingCode,
				    gateDuration, rWaveOffset,
				    numAcceptedBeats, filterCutoffFrequency,
				    filterResolution, filterRampSlope,
				    filterOrder, filterScatterFraction,
				    filterScatterSlope, annotation,
				    mt11, mt12, mt13, mt21,
				    mt22, mt23, mt31, mt32,
				    mt33, rFilterCutoff,
				    rFilterResolution, rFilterCode,
				    rFilterOrder, zFilterCutoff,
				    zFilterResolution, zFilterCode,
				    zFilterOrder, mt14, mt24,
				    mt34, scatterType, reconType,
				    reconViews, fillCti, fillUser);
    }
}
