/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.ecat;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Iterator;
import javax.imageio.stream.ImageOutputStream;

/**
 * Encoder for ECAT elements in the ECAT 7 format.
 *
 * @version 8 April 2011
 */
public class EcatElementEncoder
{
    /** Output stream for the elements. */
    private ImageOutputStream _imageStream;

    /**
     * Constructs an ECAT Element Encoder.
     *
     * @param imageStream Output stream for the elements.
     */
    public EcatElementEncoder(ImageOutputStream imageStream) 
    {
	_imageStream = imageStream;
    }

    /**
     * Writes the elements.
     *
     * @param map ECAT Element Map to encode.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void writeElements(EcatElementMap map) throws IOException
    {
	ByteArrayOutputStream bStream = new ByteArrayOutputStream(512);
	DataOutputStream outputStream = new DataOutputStream(bStream);

	// Encode each element
	Iterator iter = map.getElementNames();
	while ( iter.hasNext() ) {
	    String name = (String)iter.next();
	    Object value = map.getElementValue(name);

	    // String value
	    if (value instanceof String) {
		int size = 0;

		// ECAT 7 Main Header string elements
		if ( Ecat7MainHeader.MAGIC_NUMBER.equals(name) ) {
		    size = 14;
		}
		else if ( Ecat7MainHeader.ORIGINAL_FILE_NAME.equals(name) ) {
		    size = 32;
		}
		else if ( Ecat7MainHeader.SERIAL_NUMBER.equals(name) ) {
		    size = 10;
		}
		else if ( Ecat7MainHeader.ISOTOPE_NAME.equals(name) ) {
		    size = 8;
		}
		else if ( Ecat7MainHeader.RADIOPHARMACEUTICAL.equals(name) ) {
		    size = 32;
		}
		else if ( Ecat7MainHeader.STUDY_TYPE.equals(name) ) {
		    size = 12;
		}
		else if ( Ecat7MainHeader.PATIENT_ID.equals(name) ) {
		    size = 16;
		}
		else if ( Ecat7MainHeader.PATIENT_NAME.equals(name) ) {
		    size = 32;
		}
		else if ( Ecat7MainHeader.PATIENT_SEX.equals(name) ) {
		    size = 1;
		}
		else if ( Ecat7MainHeader.PATIENT_DEXTERITY.equals(name) ) {
		    size = 1;
		}
		else if ( Ecat7MainHeader.PHYSICIAN_NAME.equals(name) ) {
		    size = 32;
		}
		else if ( Ecat7MainHeader.OPERATOR_NAME.equals(name) ) {
		    size = 32;
		}
		else if ( Ecat7MainHeader.STUDY_DESCRIPTION.equals(name) ) {
		    size = 32;
		}
		else if ( Ecat7MainHeader.FACILITY_NAME.equals(name) ) {
		    size = 20;
		}
		else if ( Ecat7MainHeader.USER_PROCESS_CODE.equals(name) ) {
		    size = 10;
		}
		else if ( Ecat7MainHeader.DATA_UNITS.equals(name) ) {
		    size = 32;
		}

		// ECAT 7 Image Header string elements
		else if ( Ecat7ImageHeader.ANNOTATION.equals(name) ) {
		    size = 40;
		}

		// Write the string value
		_writeString( (String)value, size, outputStream );
	    }

	    // Value is a byte array
	    else if (value instanceof byte[]) {
		int size = 0;

		// ECAT 7 Image Header string elements
		if ( Ecat7ImageHeader.FILL_CTI.equals(name) ) {
		    size = 2*87;
		}
		else if ( Ecat7ImageHeader.FILL_USER.equals(name) ) {
		    size = 2*49;
		}

		// Write the byte array
		_writeByteArray( (byte[])value, size, outputStream );
	    }

	    // Element value is a Number
	    else {

		// Don't write the matrix id (it is written in a separate blk)
		if ( !Ecat7ImageHeader.MATRIX_ID.equals(name) ) {
		    _writeNumber( (Number)value, outputStream );
		}
	    }
	}

	// Write a byte array that contains the encoded Element Map
	while (bStream.size() < 512) { bStream.write(0); }
	_imageStream.write(bStream.toByteArray());
    }

    /**
     * Writes the byte array to the output stream.
     *
     * @param array Byte array to write to the output stream.
     * @param length Length of the array to actually write.
     * @param outputStream Data Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeByteArray(byte[] array, int length,
				 DataOutputStream outputStream)
	throws IOException
    {
	// Create a byte buffer of the specified length
	byte[] buffer = new byte[length];

	// Copy the specified bytes to the byte buffer
	int min = Math.min( buffer.length, array.length );
	System.arraycopy(array, 0, buffer, 0, min);

	// Write the bytes
	outputStream.write(buffer);
    }

    /**
     * Writes the Number to the output stream.
     *
     * @param number Number to write to the output stream.
     * @param outputStream Data Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeNumber(Number number, DataOutputStream outputStream)
	throws IOException
    {
	// Integer
	if (number instanceof Integer) {
	    outputStream.writeInt( ((Integer)number).intValue() );
	}

	// Short
	else if (number instanceof Short) {
	    outputStream.writeShort( ((Short)number).shortValue() );
	}

	// Float
	else if (number instanceof Float) {
	    outputStream.writeFloat( ((Float)number).floatValue() );
	}

	// Unknown Number
	else {
	    String msg = "Unable to write an Number of type \"" +
		number.getClass().getName() + "\".";
	    throw new IOException(msg);
	}
    }

    /**
     * Writes the String to the output stream.
     *
     * @param string String to write to the output stream.
     * @param length Length of the string to actually write.
     * @param outputStream Data Output Stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private void _writeString(String string, int length,
			      DataOutputStream outputStream)
	throws IOException
    {
	// Create a byte buffer of the specified length
	byte[] buffer = new byte[length];

	// Copy the bytes of the String to the byte buffer
	byte[] stringBytes = string.getBytes();
	int min = Math.min( buffer.length, stringBytes.length );
	System.arraycopy(stringBytes, 0, buffer, 0, min);

	// Write the bytes
	outputStream.write(buffer);
    }
}
