/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer;

/**
 * MGH Volume Header with RAS (Right, Anterior, Superior) information.
 *
 * @version 19 March 2009
 */
public class MghRasVolumeHeader extends MghVolumeHeader
{
    /** Voxel spacing. */
    private float[] _spacing;

    /** RAS coordinate system. */
    private float[] _ras;

    /**
     * Constructs an MGH RAS Volume Header.
     *
     * @param version Version of the header.
     * @param width Width of each image (in voxels).
     * @param height Height of each image (in voxels).
     * @param depth Number of images in a volume.
     * @param nframes Number of scalar components per voxel.
     * @param type Data type of the image.
     * @param dof Degrees of freedom.
     * @param spacingX Voxel spacing in the X direction.
     * @param spacingY Voxel spacing in the X direction.
     * @param spacingZ Voxel spacing in the X direction.
     * @param xR Component of the X axis along the right direction.
     * @param xA Component of the X axis along the anterior direction.
     * @param xS Component of the X axis along the superior direction.
     * @param yR Component of the Y axis along the right direction.
     * @param yA Component of the Y axis along the anterior direction.
     * @param yS Component of the Y axis along the superior direction.
     * @param zR Component of the Z axis along the right direction.
     * @param zA Component of the Z axis along the anterior direction.
     * @param zS Component of the Z axis along the superior direction.
     * @param cR Right coordinate value of the volume center.
     * @param cA Anterior coordinate value of the volume center.
     * @param cS Superior coordinate value of the volume center.
     */
    public MghRasVolumeHeader(int version, int width, int height, int depth,
			      int nframes, int type, int dof, float spacingX,
			      float spacingY, float spacingZ, float xR,
			      float xA, float xS, float yR, float yA, float yS,
			      float zR, float zA, float zS, float cR, float cA,
			      float cS)
    {
	super(version, width, height, depth, nframes, type, dof);

	// Voxel spacing
	_spacing = new float[3];
	_spacing[0] = spacingX;
	_spacing[1] = spacingY;
	_spacing[2] = spacingZ;

	// RAS
	_ras = new float[12];
	_ras[0] = xR;
	_ras[1] = xA;
	_ras[2] = xS;

	_ras[3] = yR;
	_ras[4] = yA;
	_ras[5] = yS;

	_ras[6] = zR;
	_ras[7] = zA;
	_ras[8] = zS;

	_ras[9] = cR;
	_ras[10] = cA;
	_ras[11] = cS;
    }

    /**
     * Gets the voxel spacing in the X direction.
     *
     * @return Voxel spacing in the X direction.
     */
    public float getSpacingX()
    {
	return _spacing[0];
    }

    /**
     * Gets the voxel spacing in the Y direction.
     *
     * @return Voxel spacing in the Y direction.
     */
    public float getSpacingY()
    {
	return _spacing[1];
    }

    /**
     * Gets the voxel spacing in the Z direction.
     *
     * @return Voxel spacing in the Z direction.
     */
    public float getSpacingZ()
    {
	return _spacing[2];
    }

    /**
     * Gets the component of the X axis along the right direction.
     *
     * @return Component of the X axis along the right direction.
     */
    public float getXr()
    {
	return _ras[0];
    }

    /**
     * Gets the component of the X axis along the anterior direction.
     *
     * @return Component of the X axis along the anterior direction.
     */
    public float getXa()
    {
	return _ras[1];
    }

    /**
     * Gets the component of the X axis along the superior direction.
     *
     * @return Component of the X axis along the superior direction.
     */
    public float getXs()
    {
	return _ras[2];
    }

    /**
     * Gets the component of the Y axis along the right direction.
     *
     * @return Component of the Y axis along the right direction.
     */
    public float getYr()
    {
	return _ras[3];
    }

    /**
     * Gets the component of the Y axis along the anterior direction.
     *
     * @return Component of the Y axis along the anterior direction.
     */
    public float getYa()
    {
	return _ras[4];
    }

    /**
     * Gets the component of the Y axis along the superior direction.
     *
     * @return Component of the Y axis along the superior direction.
     */
    public float getYs()
    {
	return _ras[5];
    }

    /**
     * Gets the component of the Z axis along the right direction.
     *
     * @return Component of the Z axis along the right direction.
     */
    public float getZr()
    {
	return _ras[6];
    }

    /**
     * Gets the component of the Z axis along the anterior direction.
     *
     * @return Component of the Z axis along the anterior direction.
     */
    public float getZa()
    {
	return _ras[7];
    }

    /**
     * Gets the component of the Z axis along the superior direction.
     *
     * @return Component of the Z axis along the superior direction.
     */
    public float getZs()
    {
	return _ras[8];
    }

    /**
     * Gets the right coordinate value of the volume center.
     *
     * @return Right coordinate value of the volume center.
     */
    public float getCr()
    {
	return _ras[9];
    }

    /**
     * Gets the anterior coordinate value of the volume center.
     *
     * @return Anterior coordinate value of the volume center.
     */
    public float getCa()
    {
	return _ras[10];
    }

    /**
     * Gets the superior coordinate value of the volume center.
     *
     * @return Superior coordinate value of the volume center.
     */
    public float getCs()
    {
	return _ras[11];
    }
}
