/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.blobio.BlobWriter;
import edu.ucla.loni.freesurfer.CurvMeta;
import java.awt.image.DataBufferByte;
import java.io.IOException;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageOutputStream;
import org.w3c.dom.Node;

/**
 * Blob Writer for the Free Surfer "New" Curv Format.
 *
 * @version 4 February 2009
 */
public class CurvBlobWriter extends BlobWriter
{
    /** Metadata decoded from the input source. */
    private CurvMeta _meta;

    /**
     * Constructs a Curv Blob Writer.
     *
     * @param origProvider Originating provider that instantiated this Object.
     */
    public CurvBlobWriter(CurvBlobWriterSpi origProvider)
    {
	super(origProvider);
    }

    /**
     * Gets an IIOMetadata object containing default values for writing a
     * series of blobs.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultStreamMetadata()
    {
	return new CurvMetadata(null);
    }

    /**
     * Gets an IIOMetadata object containing default values for writing a
     * blob.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultBlobMetadata()
    {
	return null;
    }

    /**
     * Prepares an output stream to accept a series of subsequent blob write
     * calls using the provided stream metadata object.
     *
     * @param streamMetadata IIOMetadata object representing stream metadata, or
     *                       null if none is applicable.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalStateException If the output has not been set.
     */
    public void prepareWriteBlobSequence(IIOMetadata streamMetadata)
	throws IOException
    {
	// Must have stream metadata
	if (streamMetadata == null) {
	    throw new IOException("No stream metadata.");
	}

	// Convert the metadata tree
	String rootName = CurvFormat.NATIVE_METADATA_FORMAT_NAME;
	Node tree = streamMetadata.getAsTree(rootName);
	_meta = FreeSurferMetadataConversions.toCurvMeta(tree);

	// Get the output stream
	ImageOutputStream imageStream = _getOutputStream();

	// Write the magic number:  -1 & 0x00ffffff
	byte[] magicBytes = new byte[]{(byte)0xFF,(byte)0xFF,(byte)0xFF};
	imageStream.write(magicBytes);

	// Write the number of vertices
	imageStream.writeInt( _meta.getVertexCount() );

	// Write the number of faces
	imageStream.writeInt( _meta.getFaceCount() );

	// Write the number of values per vertex
	imageStream.writeInt( _meta.getValsPerVertex() );
    }

    /**
     * Appends a single blob and possibly associated metadata to the output.
     *
     * @param blobMetadata IIOMetadata object representing blob metadata, or
     *                     null if none is applicable.
     * @param blob DataBuffer containing the binary object.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the blob is null.
     * @throws IllegalStateException If the output has not been set, or a
     *                               prepareWrite method has not been called.
     */
    public void writeToBlobSequence(IIOMetadata blobMetadata,
				    DataBufferByte blob) throws IOException
    {
	// Check for stream metadata
	if (_meta == null) {
	    String msg = "The blob sequence has not been prepared.";
	    throw new IllegalStateException(msg);
	}

	// Check for a null blob
	if (blob == null) {
	    throw new IllegalStateException("Null blob");
	}

	// Write the blob data
	ImageOutputStream imageStream = _getOutputStream();
	imageStream.write( blob.getData() );
    }

    /**
     * Completes the writing of a series of blobs begun with
     * prepareWriteBlobSequence.  Any blob metadata that should come at the
     * end of the series is written out, and any header information at the
     * beginning of the series is patched up if necessary.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalStateException If the output has not been set, or if
     *                               prepareWriteBlobSequence has not been
     *                               called.
     */
    public void endWriteBlobSequence() throws IOException
    {
	// Reset the metadata
	_meta = null;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Blob Writer.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the output stream
	try { if (_output != null) { ((ImageOutputStream)_output).close(); } }
	catch (Exception e) {}

	_output = null;
    }

    /**
     * Gets the output stream.
     *
     * @return Image Output Stream to write to.
     *
     * @throws IllegalStateException If the output has not been set.
     */
    private ImageOutputStream _getOutputStream()
    {
	// No output has been set
	if (_output == null) {
	    String msg = "No output has been set.";
	    throw new IllegalStateException(msg);
	}

	// Return the Image Output Stream for the metadata
	return (ImageOutputStream)_output;
    }
}
