/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.freesurfer.MghVolumeHeader;
import edu.ucla.loni.freesurfer.MghVolumeHeaderDecoder;
import edu.ucla.loni.imageio.Utilities;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import javax.imageio.ImageReader;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageReaderSpi;
import javax.imageio.stream.ImageInputStream;

/**
 * Image Reader for parsing and decoding MGH images.
 *
 * @version 8 April 2011
 */
public class MghImageReader extends ImageReader
{
    /** MGH Volume Header decoded from the input source. */
    private MghVolumeHeader _mghHeader;

    /**
     * Constructs an MGH Image Reader.
     *
     * @param originatingProvider The Image Reader SPI that instantiated this
     *                            Object.
     *
     * @throws IllegalArgumentException If the originating provider is not a
     *                                  Analyze Image Reader Spi.
     */
    public MghImageReader(ImageReaderSpi originatingProvider)
    {
	super(originatingProvider);

	// Originating provider must be an MGH Image Reader Spi
	if ( !(originatingProvider instanceof MghImageReaderSpi) ) {
	    String msg = "Originating provider must be an MGH Image " +
		"Reader SPI.";
	    throw new IllegalArgumentException(msg);
	}
    }

    /**
     * Sets the input source to use to the given ImageInputStream or other
     * Object.  The input source must be set before any of the query or read
     * methods are used.  If the input is null, any currently set input source
     * will be removed.
     *
     * @param input The ImageInputStream or other Object to use for future
     *              decoding.
     * @param seekForwardOnly If true, images and metadata may only be read in
     *                        ascending order from this input source.
     * @param ignoreMetadata If true, metadata may be ignored during reads.
     *
     * @throws IllegalArgumentException If the input is not an instance of one
     *                                  of the classes returned by the
     *                                  originating service provider's
     *                                  getInputTypes method, or is not an
     *                                  ImageInputStream.
     * @throws IllegalStateException If the input source is an Input Stream that
     *                               doesn't support marking and seekForwardOnly
     *                               is false.
     */
    public void setInput(Object input, boolean seekForwardOnly,
			 boolean ignoreMetadata)
    {
	if (input == null) {
	    super.setInput(input, seekForwardOnly, ignoreMetadata);
	    return;
	}

	// Convert the input source into a GZIP Image Input Stream
	try { input = new GZipImageInputStream((ImageInputStream)input); }
	catch (Exception e) {
	    String msg = "Unable to set input data";
	    IllegalArgumentException iae = new IllegalArgumentException(msg);
	    iae.initCause(e);
	    throw iae;
	}

	super.setInput(input, seekForwardOnly, ignoreMetadata);

	// Remove any decoded MGH Header
	_mghHeader = null;
    }

    /**
     * Returns the number of images, not including thumbnails, available from
     * the current input source.
     * 
     * @param allowSearch If true, the true number of images will be returned
     *                    even if a search is required.  If false, the reader
     *                    may return -1 without performing the search.
     *
     * @return The number of images or -1 if allowSearch is false and a search
     *         would be required.
     *
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     * @throws IllegalStateException If the input source has not been set.
     */
    public int getNumImages(boolean allowSearch) throws IOException
    {
	MghMetadata metadata = new MghMetadata( _getMghHeader() );
	return metadata.getNumberOfImages();
    }

    /**
     * Returns the width in pixels of the given image within the input source.
     *
     * @param imageIndex The index of the image to be queried.
     *
     * @return The width of the image in pixels.
     *
     * @throws IOException If an error occurs reading the width information from
     *                     the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public int getWidth(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	_checkImageIndex(imageIndex);

	// Return the image width
	MghMetadata metadata = new MghMetadata( _getMghHeader() );
	return metadata.getImageWidth();
    }

    /**
     * Returns the height in pixels of the given image within the input source.
     *
     * @param imageIndex The index of the image to be queried.
     *
     * @return The height of the image in pixels.
     *
     * @throws IOException If an error occurs reading the height information
     *                     from the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public int getHeight(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	_checkImageIndex(imageIndex);

	// Return the image height
	MghMetadata metadata = new MghMetadata( _getMghHeader() );
	return metadata.getImageHeight();
    }

    /**
     * Returns an Iterator containing possible image types to which the given
     * image may be decoded, in the form of ImageTypeSpecifiers.  At least one
     * legal image type will be returned.
     *
     * @param imageIndex The index of the image to be retrieved.
     *
     * @return An Iterator containing at least one ImageTypeSpecifier
     *         representing suggested image types for decoding the current given
     *         image.
     *
     * @throws IOException If an error occurs reading the format information
     *                     from the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public Iterator getImageTypes(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	_checkImageIndex(imageIndex);

	// Get the image parameters
	MghMetadata metadata = new MghMetadata( _getMghHeader() );
	int width = getWidth(imageIndex);
	int height = getHeight(imageIndex);
	int bitsPerPixel = metadata.getBitsPerPixel();
	int type = metadata.getDataType();

	// Determine the appropriate Image Type Specifier
	ArrayList list = new ArrayList(1);
	if (bitsPerPixel == 8) {
	    list.add( Utilities.getGrayImageType(width, height, bitsPerPixel) );
	}
	if (bitsPerPixel == 16) {
	    list.add( Utilities.getGrayImageType(DataBuffer.TYPE_SHORT) );
	}
	else if (type == 1) {
	    list.add( Utilities.getGrayImageType(DataBuffer.TYPE_INT) );
	}
	else if (type == 3 || type == 6) {
	    list.add( Utilities.getGrayImageType(DataBuffer.TYPE_FLOAT) );
	}

	// Check if the image data type is supported
	if (list.size() == 0) {
	    String msg = "Unsupported MGH image type:  data type = " +
		type + "; bits per voxel = " + bitsPerPixel + ".";
	    throw new IOException(msg);
	}

	// Return the Image Type Specifier
	return list.iterator();
    }

    /**
     * Returns an IIOMetadata object representing the metadata associated with
     * the input source as a whole (i.e., not associated with any particular
     * image), or null if the reader does not support reading metadata, is set
     * to ignore metadata, or if no metadata is available.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     */
    public IIOMetadata getStreamMetadata() throws IOException
    {
	return new MghMetadata( _getMghHeader() );
    }

    /**
     * Returns an IIOMetadata object containing metadata associated with the
     * given image, or null if the reader does not support reading metadata, is
     * set to ignore metadata, or if no metadata is available.
     *
     * @param imageIndex Index of the image whose metadata is to be retrieved. 
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public IIOMetadata getImageMetadata(int imageIndex) throws IOException
    {
	// Not applicable
	return null;
    }

    /**
     * Reads the image indexed by imageIndex and returns it as a complete
     * BufferedImage.  The supplied ImageReadParam is ignored.
     *
     * @param imageIndex The index of the image to be retrieved.
     * @param param An ImageReadParam used to control the reading process, or
     *              null.
     *
     * @return The desired portion of the image as a BufferedImage.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public BufferedImage read(int imageIndex, ImageReadParam param)
	throws IOException
    {
	int width = getWidth(imageIndex);
	int height = getHeight(imageIndex);

	// Update the Listeners
	processImageStarted(imageIndex);

	// Compute the size of the image in bytes
	MghMetadata metadata = new MghMetadata( _getMghHeader() );
	ImageTypeSpecifier spec = (ImageTypeSpecifier)getImageTypes(0).next();
	int dataType = spec.getSampleModel().getDataType();
	int imageSize = width*height;

	if (dataType == DataBuffer.TYPE_SHORT) { imageSize *= 2; }
	else if (dataType == DataBuffer.TYPE_INT) { imageSize *= 4; }
	else if (dataType == DataBuffer.TYPE_FLOAT) { imageSize *= 4; }

	// Seek to the indexed image (header = 284 bytes)
	ImageInputStream imageStream = (ImageInputStream)input;
	imageStream.seek(284 + imageIndex*imageSize);

	// Return the raw image data if required
	if ( Utilities.returnRawBytes(param) ) {
	    BufferedImage bufferedImage = Utilities.getRawByteImage(imageSize);
	    DataBuffer db = bufferedImage.getRaster().getDataBuffer();
	    imageStream.readFully( ((DataBufferByte)db).getData() );
	    return bufferedImage;
	}

	// Create a Buffered Image for the image data
	Iterator iter = getImageTypes(0);
	BufferedImage bufferedImage = getDestination(null, iter, width, height);
	DataBuffer dataBuffer = bufferedImage.getRaster().getDataBuffer();

	// Read the image data into an array
	byte[] srcBuffer = new byte[imageSize];
	imageStream.readFully(srcBuffer);

	// 8-bit gray image pixel data
	WritableRaster dstRaster = bufferedImage.getRaster();
	if (dataBuffer.getDataType() == DataBuffer.TYPE_BYTE) {
	    bufferedImage = Utilities.getGrayByteImage(srcBuffer, dstRaster);
	}

	// Signed 16-bit gray image pixel data
	else if (dataBuffer.getDataType() == DataBuffer.TYPE_SHORT) {
	    bufferedImage = Utilities.getGrayShortImage(srcBuffer, dstRaster,
							false);
	}
	// General gray image pixel data
	else {
	    bufferedImage = Utilities.getGrayImage(srcBuffer, dstRaster, false);
	}

	// Update the Listeners
	if ( abortRequested() ) { processReadAborted(); }
	else { processImageComplete(); }

	// Return the requested image
	return bufferedImage;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this ImageReader.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the input stream
	try { if (input != null) { ((ImageInputStream)input).close(); } }
	catch (Exception e) {}

	input = null;
    }

    /**
     * Reads the MGH Volume Header from the input source.  If the MGH Header
     * has already been decoded, no additional read is performed.
     *
     * @return MGH Volume Header decoded from the input source.
     * 
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     */
    private MghVolumeHeader _getMghHeader() throws IOException
    {
	// MGH Header already decoded
	if (_mghHeader != null) { return _mghHeader; }

	// Get the input stream
	ImageInputStream imageStream = (ImageInputStream)input;

	// Decode the header
	MghVolumeHeaderDecoder decoder=new MghVolumeHeaderDecoder(imageStream);
	_mghHeader = decoder.getHeader();
	return _mghHeader;
    }

    /**
     * Checks whether or not the specified image index is valid.
     *
     * @param imageIndex The index of the image to check.
     *
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    private void _checkImageIndex(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	if ( imageIndex >= getNumImages(false) ) {
	    String msg = "Image index of " + imageIndex + " >= the total " +
		"number of images (" + getNumImages(false) + ")";
	    throw new IndexOutOfBoundsException(msg);
	}
    }
}
