/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.freesurfer.MghVolumeHeader;
import edu.ucla.loni.freesurfer.MghVolumeHeaderEncoder;
import edu.ucla.loni.imageio.FileImageOutputStream;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.GZIPOutputStream;
import javax.imageio.IIOImage;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageWriterSpi;
import javax.imageio.stream.ImageOutputStream;
import org.w3c.dom.Node;

/**
 * Image Writer for parsing and encoding MGH images.
 *
 * @version 8 April 2011
 */
public class MghImageWriter extends ImageWriter
{
    /** MGH Header of the image sequence currently being written. */
    private MghVolumeHeader _header;

    /** Number of images in the sequence currently being written. */
    private int _numberOfImagesWritten;

    /**
     * Constructs an MGH Image Writer.
     *
     * @param originatingProvider The ImageWriterSpi that instantiated this
     *                            Object.
     *
     * @throws IllegalArgumentException If the originating provider is not an
     *                                  MGH Image Writer Spi.
     */
    public MghImageWriter(ImageWriterSpi originatingProvider)
    {
	super(originatingProvider);

	// Originating provider must be an MGH Image Writer Spi
	if ( !(originatingProvider instanceof MghImageWriterSpi) ) {
	    String msg = "Originating provider must an MGH Image Writer SPI.";
	    throw new IllegalArgumentException(msg);
	}
    }

    /**
     * Sets the destination to the given ImageOutputStream or other Object.  The
     * destination is assumed to be ready to accept data, and will not be closed
     * at the end of each write. If output is null, any currently set output
     * will be removed.
     *
     * @param output A Dual Image Output Stream or a Vector of two Image Output
     *               Streams to use for future writing.
     *
     * @throws IllegalArgumentException If output is not an instance of one of
     *                                  the classes returned by the originating
     *                                  service provider's getOutputTypes
     *                                  method.
     */
    public void setOutput(Object output)
    {
	if (output instanceof ImageOutputStream) {
	    ImageOutputStream ios = (ImageOutputStream)output;

	    // Convert to an Output Stream
	    OutputStream os = new OutputStreamAdapter(ios);

	    // File name specifies compression
	    if (output instanceof FileImageOutputStream) {
		FileImageOutputStream fStream = (FileImageOutputStream)output;
		String fileName = fStream.getFile().getName();
		if ( fileName.toLowerCase().endsWith("gz") ) {
		    try { os = new GZIPOutputStream(os); }
		    catch (Exception e) {
			String msg = "Unable to construct GZIP output stream.";
			RuntimeException re = new RuntimeException(msg);
			re.initCause(e);
			throw re;
		    }
		}
	    }
	    this.output = os;
	}

	else { super.setOutput(output); }
    }

    /**
     * Returns an IIOMetadata object containing default values for encoding a
     * stream of images.
     *
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultStreamMetadata(ImageWriteParam param)
    {
	// Return a blank metadata object
	return new MghMetadata(null);
    }

    /**
     * Returns an IIOMetadata object containing default values for encoding an
     * image of the given type.
     *
     * @param imageType An ImageTypeSpecifier indicating the format of the image
     *                  to be written later.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultImageMetadata(ImageTypeSpecifier imageType,
					       ImageWriteParam param)
    {
	// Not applicable
	return null;
    }

    /**
     * Returns an IIOMetadata object that may be used for encoding and
     * optionally modified using its document interfaces or other interfaces
     * specific to the writer plug-in that will be used for encoding.
     *
     * @param inData An IIOMetadata object representing stream metadata, used to
     *               initialize the state of the returned object.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object, or null if the plug-in does not provide
     *         metadata encoding capabilities.
     *
     * @throws IllegalArgumentException If inData is null.
     */
    public IIOMetadata convertStreamMetadata(IIOMetadata inData,
					     ImageWriteParam param)
    {
	// Only recognize MGH Metadata
	if (inData instanceof MghMetadata) { return inData; }

	// Otherwise perform no conversion
	return null;
    }

    /**
     * Returns an IIOMetadata object that may be used for encoding and
     * optionally modified using its document interfaces or other interfaces
     * specific to the writer plug-in that will be used for encoding.
     *
     * @param inData An IIOMetadata object representing image metadata, used to
     *               initialize the state of the returned object.
     * @param imageType An ImageTypeSpecifier indicating the layout and color
     *                  information of the image with which the metadata will be
     *                  associated.
     * @param param An ImageWriteParam that will be used to encode the image, or
     *              null.
     *
     * @return An IIOMetadata object, or null if the plug-in does not provide
     *         metadata encoding capabilities.
     *
     * @throws IllegalArgumentException If either of inData or imageType is
     *                                  null.
     */
    public IIOMetadata convertImageMetadata(IIOMetadata inData,
					    ImageTypeSpecifier imageType,
					    ImageWriteParam param)
    {
	// Not applicable
	return null;
    }

    /**
     * Returns true if the methods that take an IIOImage parameter are capable
     * of dealing with a Raster (as opposed to RenderedImage) source image.
     *
     * @return True if Raster sources are supported.
     */
    public boolean canWriteRasters()
    {
	return true;
    }

    /**
     * Appends a complete image stream containing a single image and associated
     * stream and image metadata and thumbnails to the output.
     *
     * @param streamMetadata An IIOMetadata object representing stream metadata,
     *                       or null to use default values.
     * @param image An IIOImage object containing an image, thumbnails, and
     *              metadata to be written.
     * @param param An ImageWriteParam, or null to use a default
     *              ImageWriteParam.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the image is null, if the stream
     *                                  metadata cannot be converted, or if the
     *                                  image type is unrecognized.
     * @throws IllegalStateException If the output has not been set.
     */
    public void write(IIOMetadata streamMetadata, IIOImage image,
		      ImageWriteParam param) throws IOException
    {
	// Write a sequence with 1 image
	prepareWriteSequence(streamMetadata);
	writeToSequence(image, param);
	endWriteSequence();
    }

    /**
     * Returns true if the writer is able to append an image to an image
     * stream that already contains header information and possibly prior
     * images.
     *
     * @return True If images may be appended sequentially.
     */
    public boolean canWriteSequence()
    {
	return true;
    }

    /**
     * Prepares a stream to accept a series of subsequent writeToSequence calls,
     * using the provided stream metadata object.  The metadata will be written
     * to the stream if it should precede the image data.
     *
     * @param streamMetadata A stream metadata object.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the stream metadata is null or cannot
     *                                  be converted.
     * @throws IllegalStateException If the output has not been set.
     */
    public void prepareWriteSequence(IIOMetadata streamMetadata)
	throws IOException
    {
	// Convert the metadata to MGH Metadata
	IIOMetadata metadata = convertStreamMetadata(streamMetadata, null);

	// Unable to convert the metadata
	if (metadata == null) {
	    String msg = "Unable to convert the stream metadata for encoding.";
	    throw new IllegalArgumentException(msg);
	}

	// Construct a stream MGH Header from the metadata
	try {
	    String rootName = MghMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	    Node tree = metadata.getAsTree(rootName);
	    _header = FreeSurferMetadataConversions.toMghVolumeHeader(tree);
	}
      
	catch (Exception e) {
	    String msg = "Unable to construct an MGH Header from the metadata.";
	    throw new IllegalArgumentException(msg);
	}

	// No output has been set
	if (output == null) {
	    String msg = "No output has been set.";
	    throw new IllegalStateException(msg);
	}

	// Write the MGH header
	DataOutputStream dos = new DataOutputStream((OutputStream)output);
	MghVolumeHeaderEncoder encoder = new MghVolumeHeaderEncoder(dos);
	encoder.writeMainElements(_header);

	_numberOfImagesWritten = 0;
    }

    /**
     * Appends a single image and possibly associated metadata and thumbnails,
     * to the output.  The supplied ImageReadParam is ignored.
     *
     * @param image An IIOImage object containing an image, thumbnails, and
     *              metadata to be written.
     * @param param An ImageWriteParam or null to use a default ImageWriteParam.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the image is null or if the image
     *                                  type is not recognized.
     * @throws IllegalStateException If the output has not been set, or
     *                               prepareWriteSequence has not been called.
     */
    public void writeToSequence(IIOImage image, ImageWriteParam param)
	throws IOException
    {
	// Check for the MGH Header
	if (_header == null) {
	    String msg = "The image sequence has not been prepared.";
	    throw new IllegalStateException(msg);
	}

	// Check for a null image
	if (image == null) {
	    String msg = "Cannot write a null image.";
	    throw new IllegalArgumentException(msg);
	}

	// Get a Raster from the image
	Raster raster = image.getRaster();
	if (raster == null) {
	    RenderedImage renderedImage = image.getRenderedImage();

	    // If the Rendered Image is a Buffered Image, get it directly
	    if (renderedImage instanceof BufferedImage) {
		raster = ((BufferedImage)renderedImage).getRaster();
	    }

	    // Otherwise get a copy of the Raster from the Rendered Image
	    else { raster = renderedImage.getData(); }
	}
	DataBuffer dataBuffer = raster.getDataBuffer();

	// Update the Listeners
	processImageStarted(_numberOfImagesWritten);

	// Write byte data
	OutputStream os = (OutputStream)output;
	if (dataBuffer instanceof DataBufferByte) {
	    os.write( ((DataBufferByte)dataBuffer).getData() );
	}

	// Write short data
	else if (dataBuffer instanceof DataBufferShort) {
	    short[] data = ((DataBufferShort)dataBuffer).getData();
	    _writeShorts(data, os);
	}

	// Write integer data
	else if (dataBuffer instanceof DataBufferInt) {
	    int[] data = ((DataBufferInt)dataBuffer).getData();
	    _writeInts(data, os);
	}

	// Write float data
	else if (dataBuffer instanceof DataBufferFloat) {
	    float[] data = ((DataBufferFloat)dataBuffer).getData();
	    _writeFloats(data, os);
	}

	// Unrecognized type
	else {
	    String msg = "Unable to write the IIOImage.";
	    throw new IllegalArgumentException(msg);
	}

	// Update the Listeners
	_numberOfImagesWritten++;
	if ( abortRequested() ) { processWriteAborted(); }
	else { processImageComplete(); }
    }

    /**
     * Completes the writing of a sequence of images begun with
     * prepareWriteSequence.  Any stream metadata that should come at the end of
     * the sequence of images is written out, and any header information at the
     * beginning of the sequence is patched up if necessary.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void endWriteSequence() throws IOException
    {
	// Write the MGH footer
	DataOutputStream dos = new DataOutputStream((OutputStream)output);
	MghVolumeHeaderEncoder encoder = new MghVolumeHeaderEncoder(dos);
	encoder.writeAcquistionInfo(_header);

	// Reset the MGH Header
	_header = null;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Image Writer.  Otherwise, the writer may continue to hold
     * on to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the output stream
	try { if (output != null) { ((OutputStream)output).close(); } }
	catch (Exception e) {}

	output = null;
    }

    /**
     * Writes the shorts to the output stream.
     *
     * @param s Shorts to write.
     * @param outputStream Output stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private static void _writeShorts(short[] s, OutputStream outputStream)
	throws IOException
    {
	byte[] b = new byte[2*s.length];

	int boff = 0;
	for (int i = 0; i < s.length; i++) {
	    b[boff++] = (byte)(s[i] >>> 8);
	    b[boff++] = (byte)(s[i] >>> 0);
	}

	outputStream.write(b);
    }

    /**
     * Writes the integers to the output stream.
     *
     * @param i Integers to write.
     * @param outputStream Output stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private static void _writeInts(int[] i, OutputStream outputStream)
	throws IOException
    {
	byte[] b = new byte[4*i.length];

	int boff = 0;
	for (int j = 0; j < i.length; j++) {
	    b[boff++] = (byte)(i[j] >>> 24);
	    b[boff++] = (byte)(i[j] >>> 16);
	    b[boff++] = (byte)(i[j] >>> 8);
	    b[boff++] = (byte)(i[j] >>> 0);
	}

	outputStream.write(b);
    }

    /**
     * Writes the floats to the output stream.
     *
     * @param f Floats to write.
     * @param outputStream Output stream to write to.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private static void _writeFloats(float[] f, OutputStream outputStream)
	throws IOException
    {
	byte[] b = new byte[4*f.length];

	int boff = 0;
	for (int i = 0; i < f.length; i++) {
	    int v = Float.floatToIntBits(f[i]);
	    b[boff++] = (byte)(v >>> 24);
	    b[boff++] = (byte)(v >>> 16);
	    b[boff++] = (byte)(v >>> 8);
	    b[boff++] = (byte)(v >>> 0);
	}

	outputStream.write(b);
    }
}
