/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.freesurfer.MghRasVolumeHeader;
import edu.ucla.loni.freesurfer.MghVolumeHeader;
import edu.ucla.loni.imageio.AppletFriendlyIIOMetadata;
import edu.ucla.loni.imageio.BasicMetadataFormat;
import edu.ucla.loni.imageio.BasicMetadataTreeMaker;
import edu.ucla.loni.imageio.Utilities;
import java.util.Iterator;
import java.util.Vector;
import javax.imageio.metadata.IIOInvalidTreeException;
import javax.imageio.metadata.IIOMetadataFormatImpl;
import javax.imageio.metadata.IIOMetadataNode;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * IIO Metadata for metadata of the MGH Format.
 *
 * @version 19 March 2009
 */
public class MghMetadata extends AppletFriendlyIIOMetadata
{
    /** MGH Volume Header containing the original metadata. */
    private MghVolumeHeader _origHeader;

    /** MGH Volume Header containing the current metadata. */
    private MghVolumeHeader _currentHeader;

    /**
     * Constructs an MGH Metadata.
     *
     * @param header MGH Volume Header containing the metadata.
     */
    public MghMetadata(MghVolumeHeader header)
    {
	super(true, MghMetadataFormat.NATIVE_METADATA_FORMAT_NAME,
	      MghMetadataFormat.class.getName(),
	      new String[]{BasicMetadataFormat.NATIVE_METADATA_FORMAT_NAME},
	      new String[]{BasicMetadataFormat.class.getName()});

	_origHeader = header;
	_currentHeader = header;
    }

    /**
     * Returns true if this object does not support the mergeTree, setFromTree,
     * and reset methods.  
     *
     * @return True if this IIOMetadata object cannot be modified; false o/w.
     */
    public boolean isReadOnly()
    {
	return false;
    }

    /**
     * Returns an XML DOM Node object that represents the root of a tree of
     * metadata contained within this object according to the conventions
     * defined by a given metadata format.
     *
     * @param formatName Name of the desired metadata format.
     *
     * @return An XML DOM Node object forming the root of a tree.
     *
     * @throws IllegalArgumentException If formatName is not one of the allowed
     *                                  metadata format names.
     */
    public Node getAsTree(String formatName)
    {
	// Standard format name
	String standardName = IIOMetadataFormatImpl.standardMetadataFormatName;
	if ( standardName.equals(formatName) ) { return getStandardTree(); }

	// Native metadata format name
	if ( nativeMetadataFormatName.equals(formatName) ) {
	    return FreeSurferMetadataConversions.toTree(_currentHeader);
	}

	// LONI Basic format name
	String basicName = BasicMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( basicName.equals(formatName) ) { return _getBasicTree(); }

	// Unknown format name
	String msg = "The format name \"" + formatName + "\" is not a " +
	    "supported metadata format name.";
	throw new IllegalArgumentException(msg);
    }

    /**
     * Alters the internal state of this IIOMetadata object from a tree of XML
     * DOM Nodes whose syntax is defined by the given metadata format.  The
     * previous state is altered only as necessary to accomodate the nodes that
     * are present in the given tree.
     *
     * @param formatName Name of the desired metadata format.
     * @param root An XML DOM Node object forming the root of a tree.
     *
     * @throws IllegalArgumentException If formatName is not one of the allowed
     *                                  metadata format names, or if the root is
     *                                  null.
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     */
    public void mergeTree(String formatName, Node root)
	throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check for a matching root Node
	if ( !root.getNodeName().equals(nativeMetadataFormatName) ) {
	    String msg = "Root node must be named \"" +
		nativeMetadataFormatName + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// If there is no current Header, convert the root
	Node r;
	if (_currentHeader == null) { r = root; }

	// Otherwise merge the root with the current metadata tree Nodes
	else {
	    Node headerNode = getAsTree(formatName);
	    r = _getMergedNode(headerNode, root.getAttributes());

	    // Get the acquisition Node
	    Node acqNode = null;
	    Node childNode = headerNode.getFirstChild();
	    if (childNode != null) {
		String cName = childNode.getNodeName();
		if (MghMetadataFormat.ACQUISITION.equals(cName)) {
		    acqNode = childNode;
		}
	    }

	    // Replace the acquisition Node
	    childNode = root.getFirstChild();
	    if (childNode != null) {
		String cName = childNode.getNodeName();
		if (MghMetadataFormat.ACQUISITION.equals(cName)) {
		    acqNode = childNode;
		}
	    }

	    // Get the tag Nodes
	    Vector tagNodes = new Vector();
	    NodeList nodeList = headerNode.getChildNodes();
	    for (int i = 0; i < nodeList.getLength(); i++) {
		childNode = nodeList.item(i);
		String cName = childNode.getNodeName();
		if (MghMetadataFormat.TAG.equals(cName)) {
		    tagNodes.add(childNode);
		}
	    }

	    // Add to the tag Nodes
	    nodeList = root.getChildNodes();
	    for (int i = 0; i < nodeList.getLength(); i++) {
		childNode = nodeList.item(i);
		String cName = childNode.getNodeName();
		if (MghMetadataFormat.TAG.equals(cName)) {
		    tagNodes.add(childNode);
		}
	    }

	    // Append the acquisition Node
	    if (acqNode != null) { r.appendChild(acqNode); }

	    // Append the tag Nodes
	    Iterator iter = tagNodes.iterator();
	    while ( iter.hasNext() ) {
		Node tagNode = (Node)iter.next();
		r.appendChild(tagNode);
	    }
	}

	// Set a new current Header
	_currentHeader = FreeSurferMetadataConversions.toMghVolumeHeader(r);
    }

    /**
     * Resets all the data stored in this object to default values, usually to
     * the state this object was in immediately after construction, though the
     * precise semantics are plug-in specific.  Note that there are many
     * possible default values, depending on how the object was created.
     */
    public void reset()
    {
	// Reset the current Header to the original 
	_currentHeader = _origHeader;
    }

    /**
     * Gets the number of images.
     *
     * @return Number of images, or -1 if it cannot be determined.
     */
    public int getNumberOfImages()
    {
	return _currentHeader.getDepth()*_currentHeader.getNframes();
    }

    /**
     * Gets the width of each image.
     *
     * @return Width of each image, or -1 if it cannot be determined.
     */
    public int getImageWidth()
    {
	return _currentHeader.getWidth();
    }

    /**
     * Gets the height of each image.
     *
     * @return Height of each image, or -1 if it cannot be determined.
     */
    public int getImageHeight()
    {
	return _currentHeader.getHeight();
    }

    /**
     * Gets the number of bits per pixel used by each image.
     *
     * @return Number of bits per pixel used by each image, or -1 if it cannot
     *         be determined.
     */
    public int getBitsPerPixel()
    {
	// Get the data type
	int type = getDataType();

	// Uchar
	if (type == 0) { return 8; }

	// Integer
	if (type == 1) { return 32; }

	// Float
	if (type == 3) { return 32; }

	// Short
	if (type == 4) { return 16; }

	// Tensor = float
	if (type == 6) { return 32; }

	// Unknown type
	return -1;
    }

    /**
     * Gets the data type of each image.
     *
     * @return Data type of each image (as defined by the MGH file format),
     *         or -1 if it cannot be determined.
     */
    public int getDataType()
    {
	return _currentHeader.getType();
    }

    /**
     * Returns an IIOMetadataNode representing the chroma information of the
     * standard javax_imageio_1.0 metadata format, or null if no such
     * information is available.
     *
     * @return IIOMetadataNode representing the chroma information, or null if
     *         it does not exist.
     */
    protected IIOMetadataNode getStandardChromaNode()
    {
	// Only support non-inverted grayscale
	return Utilities.getGrayScaleChromaNode(false);
    }

    /**
     * Returns an IIOMetadataNode representing the data format information of
     * the standard javax_imageio_1.0 metadata format, or null if no such
     * information is available.
     *
     * @return IIOMetadataNode representing the data information, or null if
     *         it does not exist.
     */
    protected IIOMetadataNode getStandardDataNode()
    {
	int type = _currentHeader.getType();

	// Get the number of bits per pixel
	int bitsPerPixel = getBitsPerPixel();
	if (bitsPerPixel == -1) { return null; }

	// Integral gray scale data
	if (type == 0 || type == 1 || type == 4) {
	    return Utilities.getGrayScaleDataNode(bitsPerPixel);
	}

	// Real gray scale data
	return Utilities.getGrayScaleDataNode(bitsPerPixel, false);
    }

    /**
     * Returns an IIOMetadataNode representing the dimension format information
     * of the standard javax_imageio_1.0 metadata format, or null if no such
     * information is available.
     *
     * @return IIOMetadataNode representing the dimension information, or null
     *         if it does not exist.
     */
    protected IIOMetadataNode getStandardDimensionNode()
    {
	float pixelWidth = 1;
	float pixelHeight = 1;
	if (_currentHeader instanceof MghRasVolumeHeader) {
	    MghRasVolumeHeader rasHeader = (MghRasVolumeHeader)_currentHeader;
	    pixelWidth = rasHeader.getSpacingX();
	    pixelHeight = rasHeader.getSpacingY();
	}

	// Return the dimension node
	return Utilities.getDimensionNode(pixelWidth, pixelHeight);
    }

    /**
     * Gets the LONI Basic metadata tree.
     *
     * @return DOM Node representing the LONI Basic metadata, or null if it can
     *         not be determined.
     */
    private Node _getBasicTree()
    {
	// Image dimensions
	int imageWidth = getImageWidth();
	int imageHeight = getImageHeight();
	int numberOfImages = getNumberOfImages();

	// Pixel dimensions
	float pixelWidth = 1;
	float pixelHeight = 1;
	float sliceThickness = 1;
	if (_currentHeader instanceof MghRasVolumeHeader) {
	    MghRasVolumeHeader rasHeader = (MghRasVolumeHeader)_currentHeader;
	    pixelWidth = rasHeader.getSpacingX();
	    pixelHeight = rasHeader.getSpacingY();
	    sliceThickness = rasHeader.getSpacingZ();
	}

	// Data type
	int type = _currentHeader.getType();
	String dataType = "Unknown";
	if (type == 0) { dataType = BasicMetadataFormat.BYTE_TYPE; }
	if (type == 1) { dataType = BasicMetadataFormat.SIGNED_INT_TYPE; }
	if (type == 3) { dataType = BasicMetadataFormat.FLOAT_TYPE; }
	if (type == 4) { dataType = BasicMetadataFormat.SIGNED_SHORT_TYPE; }
	if (type == 6) { dataType = BasicMetadataFormat.FLOAT_TYPE; }

	// Bits per pixel
	int bitsPerPixel = getBitsPerPixel();

	// Color type
	String colorType = BasicMetadataFormat.GRAY_TYPE;

	// Return the basic tree
	return BasicMetadataTreeMaker.getTree(imageWidth, imageHeight,
					      numberOfImages, pixelWidth,
					      pixelHeight, sliceThickness,
					      dataType, bitsPerPixel,
					      colorType);
    }

    /**
     * Merges the attributes with the specified Node.  If there are attribute
     * names not present in the attributes of the node, the attributes are
     * ignored.
     *
     * @param node Node to merge the attributes with.
     * @param attributes Attributes to merge.
     *
     * @return New Node containing the merged attributes.
     */
    private Node _getMergedNode(Node node, NamedNodeMap attributes)
    {
	// Create a new Node with the same name
	IIOMetadataNode mergedNode = new IIOMetadataNode( node.getNodeName() );

	// Copy the Node attributes to the new Node
	NamedNodeMap map = node.getAttributes();
	if (map != null) {
	    for (int i = 0; i < map.getLength(); i++) {
		Node attr = map.item(i);
		mergedNode.setAttribute(attr.getNodeName(),attr.getNodeValue());
	    }
	}

	// Merge the new attributes with the old
	if (attributes != null) {
	    for (int i = 0; i < attributes.getLength(); i++) {
		Node attribute = attributes.item(i);
		String attributeName = attribute.getNodeName();

		// Attribute name must be present in the new Node
		if ( mergedNode.hasAttribute(attributeName) ) {
		    mergedNode.setAttribute(attributeName,
					    attribute.getNodeValue());
		}
	    }
	}

	// Return the new merged Node
	return mergedNode;
    }
}
