/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.freesurfer.MghFloat;
import edu.ucla.loni.imageio.AppletFriendlyIIOMetadataFormat;
import javax.imageio.ImageTypeSpecifier;

/**
 * IIO Metadata Format that describes the tree structure of metadata in the
 * Massachusetts General Hospital (MGH) image format.
 *
 * @version 16 March 2009
 */
public class MghMetadataFormat extends AppletFriendlyIIOMetadataFormat
{
    /** Name of the native metadata format. */
    public static final String NATIVE_METADATA_FORMAT_NAME =
	"edu_ucla_loni_mgh_1.0";

    /** Name of the acquisition Node. */
    public static final String ACQUISITION = "ACQUISITION";

    /** Name of the tag Nodes. */
    public static final String TAG = "TAG";

    /** Version of the header. */
    public static final String VERSION = "version";

    /** Width of each image (in voxels). */
    public static final String WIDTH = "width";

    /** Height of each image (in voxels). */
    public static final String HEIGHT = "height";

    /** Number of images in a volume. */
    public static final String DEPTH = "depth";

    /** Number of scalar components per voxel. */
    public static final String NFRAMES = "nframes";

    /** Data type of the image. */
    public static final String TYPE = "type";

    /** Degrees of freedom. */
    public static final String DOF = "dof";

    /** Voxel spacing in the X direction. */
    public static final String SPACING_X = "spacingX";

    /** Voxel spacing in the Y direction. */
    public static final String SPACING_Y = "spacingY";

    /** Voxel spacing in the Z direction. */
    public static final String SPACING_Z = "spacingZ";

    /** Component of the X axis along the right direction. */
    public static final String XR = "xR";

    /** Component of the X axis along the anterior direction. */
    public static final String XA = "xA";

    /** Component of the X axis along the superior direction. */
    public static final String XS = "xS";

    /** Component of the Y axis along the right direction. */
    public static final String YR = "yR";

    /** Component of the Y axis along the anterior direction. */
    public static final String YA = "yA";

    /** Component of the Y axis along the superior direction. */
    public static final String YS = "yS";

    /** Component of the Z axis along the right direction. */
    public static final String ZR = "zR";

    /** Component of the Z axis along the anterior direction. */
    public static final String ZA = "zA";

    /** Component of the Z axis along the superior direction. */
    public static final String ZS = "zS";

    /** Right coordinate value of the volume center. */
    public static final String CR = "cR";

    /** Anterior coordinate value of the volume center. */
    public static final String CA = "cA";

    /** Superior coordinate value of the volume center. */
    public static final String CS = "cS";

    /** Tag code. */
    public static final String TAG_CODE = "code";

    /** Tag value. */
    public static final String TAG_VALUE = "value";

    /** Single instance of the MGH Metadata Format. */
    private static MghMetadataFormat _format = new MghMetadataFormat();

    /** Constructs an MGH Metadata Format. */
    private MghMetadataFormat()
    {
      // Create the root with optional children
      super(NATIVE_METADATA_FORMAT_NAME, CHILD_POLICY_SOME);
      String rootName = NATIVE_METADATA_FORMAT_NAME;

      // Add the required attributes
      addAttribute(rootName, VERSION, DATATYPE_INTEGER, true, null);
      addAttribute(rootName, WIDTH, DATATYPE_INTEGER, true, null);
      addAttribute(rootName, HEIGHT, DATATYPE_INTEGER, true, null);
      addAttribute(rootName, DEPTH, DATATYPE_INTEGER, true, null);
      addAttribute(rootName, NFRAMES, DATATYPE_INTEGER, true, null);
      addAttribute(rootName, TYPE, DATATYPE_INTEGER, true, null);
      addAttribute(rootName, DOF, DATATYPE_INTEGER, true, null);
      addAttribute(rootName, SPACING_X, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, SPACING_Y, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, SPACING_Z, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, XR, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, XA, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, XS, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, YR, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, YA, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, YS, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, ZR, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, ZA, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, ZS, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, CR, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, CA, DATATYPE_FLOAT, true, null);
      addAttribute(rootName, CS, DATATYPE_FLOAT, true, null);

      // Add the acquisition child with no children
      addElement(ACQUISITION, rootName, CHILD_POLICY_EMPTY);

      // Add the acquisition attributes
      addAttribute(ACQUISITION, MghFloat.TR, DATATYPE_FLOAT, false, null);
      addAttribute(ACQUISITION, MghFloat.FLIP_ANGLE, DATATYPE_FLOAT, false,
		   null);
      addAttribute(ACQUISITION, MghFloat.TE, DATATYPE_FLOAT, false, null);
      addAttribute(ACQUISITION, MghFloat.TI, DATATYPE_FLOAT, false, null);
      addAttribute(ACQUISITION, MghFloat.FOV, DATATYPE_FLOAT, false, null);

      // Add the tag children
      addElement(TAG, rootName, CHILD_POLICY_EMPTY);
      
      // Add the tag attributes
      addAttribute(TAG, TAG_CODE, DATATYPE_INTEGER, true, null);
      addAttribute(TAG, TAG_VALUE, DATATYPE_STRING, true, null);
    }

    /**
     * Gets an instance of the MGH Metadata Format.
     *
     * @return The single instance of the MGH Metadata Format.
     */
    public static MghMetadataFormat getInstance()
    {
	return _format;
    }

    /**
     * Returns true if the element (and the subtree below it) is allowed to
     * appear in a metadata document for an image of the given type, defined by
     * an ImageTypeSpecifier.
     *
     * @param elementName The name of the element being queried.
     * @param imageType An ImageTypeSpecifier indicating the type of the image
     *                  that will be associated with the metadata.
     *
     * @return True if the node is meaningful for images of the given type.
     */
    public boolean canNodeAppear(String elementName,
				 ImageTypeSpecifier imageType)
    {
	// Not implemented
	return true;
    }
}
