/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.blobio.BlobReader;
import edu.ucla.loni.freesurfer.XfmMeta;
import java.awt.image.DataBufferByte;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageInputStream;

/**
 * Blob Reader for the Free Surfer XFM Format.
 *
 * @version 14 January 2010
 */
public class XfmBlobReader extends BlobReader
{
    /** Metadata decoded from the input source. */
    private XfmMeta _meta;

    /**
     * Constructs an XFM Blob Reader.
     *
     * @param origProvider The Blob Reader SPI that instantiated this Object.
     */
    public XfmBlobReader(XfmBlobReaderSpi origProvider)
    {
	super(origProvider);
    }

    /**
     * Sets the input source to the specified Object.  The input source must be
     * set before any of the query or read methods are used.  If the input is
     * null, any currently set input source will be removed.
     *
     * @param input Object which is the source of the input.
     *
     * @throws IllegalArgumentException If the input is not an instance of one
     *                                  of the classes returned by the
     *                                  originating service provider's
     *                                  getInputTypes method.
     */
    public void setInput(Object input)
    {
	super.setInput(input);

	// Remove any decoded metadata
	_meta = null;
    }

    /**
     * Returns the number of blobs available from the current input source.
     *
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return Number of blobs available from the current input source.
     *
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     * @throws IllegalStateException If the input source has not been set.
     */
    public int getNumberOfBlobs(int maxSize) throws IOException
    {
	// No binary data
	return 0;
    }

    /**
     * Returns an IIOMetadata object representing the metadata associated with
     * the input source as a whole (i.e., not associated with any particular
     * blob), or null if the reader does not support reading metadata, is
     * set to ignore metadata, or if no metadata is available.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     */
    public IIOMetadata getStreamMetadata() throws IOException
    {
	return new XfmMetadata( _getMeta() );
    }

    /**
     * Returns an IIOMetadata object containing metadata associated with the
     * given blob, or null if no metadata is available.
     *
     * @param blobIndex Index of the blob to retrieve.
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public IIOMetadata getBlobMetadata(int blobIndex, int maxSize)
	throws IOException
    {
	// Not applicable
	return null;
    }

    /**
     * Returns the specified binary large object (BLOB).
     *
     * @param blobIndex Index of the blob to retrieve.
     * @param maxSize Maximum size of a blob.  A blob larger than the maximum
     *                size will be subdivided into blobs smaller than this size.
     *
     * @return Binary object.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public DataBufferByte getBlob(int blobIndex, int maxSize) throws IOException
    {
	throw new IndexOutOfBoundsException("No binary data");
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Blob Reader.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the input stream
	try { if (_input != null) { ((ImageInputStream)_input).close(); } }
	catch (Exception e) {}

	_input = null;
    }

    /**
     * Gets metadata from the input source.
     *
     * @return Metadata from the input source.
     *
     * @throws IOException If an I/O error occurs.
     * @throws IllegalStateException If the input has not been set.
     */
    private XfmMeta _getMeta() throws IOException
    {
	// Metadata has already been decoded
	if (_meta != null) { return _meta; }

	// Metadata to extract
	String transformType = null;
	Vector comments = new Vector();
	String[] tranformElems = null;

	// Parse lines of text
	ImageInputStream imageStream = _getInputStream();
	String line = imageStream.readLine();
	while (line != null) {

	    // Transform type
	    if (line.indexOf("Transform_Type") != -1) {
		try {
		    StringTokenizer tokenizer = new StringTokenizer(line," =;");
		    tokenizer.nextToken();
		    transformType = tokenizer.nextToken();
		}
		catch (Exception e) {
		    String msg = "Unable to parse Transform_Type";
		    IOException ioe = new IOException(msg);
		    ioe.initCause(e);
		    throw ioe;
		}
	    }

	    // Comment
	    else if ( line.startsWith("%") ) {
		String comment = line.substring(1);
		comments.add(comment);
	    }

	    // Linear transform
	    else if (line.indexOf("Linear_Transform") != -1) {
		try {
		    tranformElems = new String[12];

		    // Line 1
		    String line1 = imageStream.readLine();
		    String[] floats = _getFourFloats(line1);
		    tranformElems[0] = floats[0];
		    tranformElems[1] = floats[1];
		    tranformElems[2] = floats[2];
		    tranformElems[3] = floats[3];

		    // Line 2
		    String line2 = imageStream.readLine();
		    floats = _getFourFloats(line2);
		    tranformElems[4] = floats[0];
		    tranformElems[5] = floats[1];
		    tranformElems[6] = floats[2];
		    tranformElems[7] = floats[3];

		    // Line 3
		    String line3 = imageStream.readLine();
		    floats = _getFourFloats(line3);
		    tranformElems[8] = floats[0];
		    tranformElems[9] = floats[1];
		    tranformElems[10] = floats[2];
		    tranformElems[11] = floats[3];
		}
		catch (Exception e) {
		    String msg = "Unable to parse Linear_Transform";
		    IOException ioe = new IOException(msg);
		    ioe.initCause(e);
		    throw ioe;
		}
	    }

	    // Read the next line
	    line = imageStream.readLine();
	}

	// Check for the transform type
	if (transformType == null) {
	    throw new IOException("Missing Transform_Type element");
	}

	// Check for the transformation matrix elements
	if (tranformElems == null) {
	    throw new IOException("Missing Linear_Transform element");
	}

	// Create the metadata
	_meta = new XfmMeta(transformType,
			    (String[])comments.toArray(new String[0]),
			    tranformElems);
	return _meta;
    }

    /**
     * Parses the string and returns 4 floats.
     *
     * @param s String to parse.
     *
     * @return Four float numbers.
     *
     * @throws IllegalArgumentException If the string does not contain 4 floats.
     * @throws NumberFormatException If the string does not contain floats.
     */
    private static String[] _getFourFloats(String s)
    {
	String[] floats = new String[4];

	// Parse the line
	StringTokenizer tokenizer = new StringTokenizer(s," ;");

	// Get four floats
	try {
	    floats[0] = tokenizer.nextToken();
	    floats[1] = tokenizer.nextToken();
	    floats[2] = tokenizer.nextToken();
	    floats[3] = tokenizer.nextToken();
	}
	catch (Exception e) {
	    String msg = "Missing 4 floats on line:  " + s;
	    IllegalArgumentException iae = new IllegalArgumentException(msg);
	    iae.initCause(e);
	    throw iae;
	}

	// Check that each is a float
	for (int i = 0; i < floats.length; i++) {
	    Float.parseFloat(floats[i]);
	}

	// Return the floats
	return floats;
    }

    /**
     * Gets the input stream.
     *
     * @return Image Input Stream to read data from.
     *
     * @throws IllegalStateException If the input has not been set.
     */
    private ImageInputStream _getInputStream()
    {
	// No input has been set
	if (_input == null) {
	    String msg = "No input has been set.";
	    throw new IllegalStateException(msg);
	}

	// Return the Image Input Stream
	return (ImageInputStream)_input;
    }
}
