/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.blobio.BlobWriter;
import edu.ucla.loni.freesurfer.XfmMeta;
import java.awt.image.DataBufferByte;
import java.io.IOException;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageOutputStream;
import org.w3c.dom.Node;

/**
 * Blob Writer for the Free Surfer XFM Format.
 *
 * @version 14 January 2010
 */
public class XfmBlobWriter extends BlobWriter
{
    /** Metadata decoded from the input source. */
    private XfmMeta _meta;

    /**
     * Constructs an XFM Blob Writer.
     *
     * @param origProvider Originating provider that instantiated this Object.
     */
    public XfmBlobWriter(XfmBlobWriterSpi origProvider)
    {
	super(origProvider);
    }

    /**
     * Gets an IIOMetadata object containing default values for writing a
     * series of blobs.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultStreamMetadata()
    {
	return new XfmMetadata(null);
    }

    /**
     * Gets an IIOMetadata object containing default values for writing a
     * blob.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultBlobMetadata()
    {
	return null;
    }

    /**
     * Prepares an output stream to accept a series of subsequent blob write
     * calls using the provided stream metadata object.
     *
     * @param streamMetadata IIOMetadata object representing stream metadata, or
     *                       null if none is applicable.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalStateException If the output has not been set.
     */
    public void prepareWriteBlobSequence(IIOMetadata streamMetadata)
	throws IOException
    {
	// Must have stream metadata
	if (streamMetadata == null) {
	    throw new IOException("No stream metadata.");
	}

	// Convert the metadata tree
	String rootName = XfmFormat.NATIVE_METADATA_FORMAT_NAME;
	Node tree = streamMetadata.getAsTree(rootName);
	_meta = FreeSurferMetadataConversions.toXfmMeta(tree);

	// Get the output stream
	ImageOutputStream imageStream = _getOutputStream();

	// Write the magic number
	imageStream.writeBytes("MNI Transform File\n");

	// Write the comments
	String[] comments = _meta.getComments();
	for (int i = 0; i < comments.length; i++) {
	    imageStream.writeBytes("%" + comments[i] + "\n" );
	}

	// Write an empty line
	imageStream.writeBytes("\n");

	// Write the transform type
	imageStream.writeBytes("Transform_Type = " + _meta.getTransformType() +
			       ";\n");

	// Write the transform matrix
	imageStream.writeBytes("Linear_Transform =\n");

	// Pad the transform matrix elements
	String[] col1 = new String[]{_meta.getTransformElem(0),
				     _meta.getTransformElem(4),
				     _meta.getTransformElem(8)};
	col1 = _pad(col1);

	String[] col2 = new String[]{_meta.getTransformElem(1),
				     _meta.getTransformElem(5),
				     _meta.getTransformElem(9)};
	col2 = _pad(col2);

	String[] col3 = new String[]{_meta.getTransformElem(2),
				     _meta.getTransformElem(6),
				     _meta.getTransformElem(10)};
	col3 = _pad(col3);

	String[] col4 = new String[]{_meta.getTransformElem(3),
				     _meta.getTransformElem(7),
				     _meta.getTransformElem(11)};
	col4 = _pad(col4);

	// Write the transform matrix elements
	for (int i = 0; i < 3; i++) {
	    imageStream.writeBytes(col1[i] + col2[i] + col3[i] + col4[i]);
	    if (i < 2) { imageStream.writeBytes("\n"); }
	    else { imageStream.writeBytes(" ;\n"); }
	}
    }

    /**
     * Appends a single blob and possibly associated metadata to the output.
     *
     * @param blobMetadata IIOMetadata object representing blob metadata, or
     *                     null if none is applicable.
     * @param blob DataBuffer containing the binary object.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the blob is null.
     * @throws IllegalStateException If the output has not been set, or a
     *                               prepareWrite method has not been called.
     */
    public void writeToBlobSequence(IIOMetadata blobMetadata,
				    DataBufferByte blob) throws IOException
    {
	throw new IOException("Cannot write blobs");
    }

    /**
     * Completes the writing of a series of blobs begun with
     * prepareWriteBlobSequence.  Any blob metadata that should come at the
     * end of the series is written out, and any header information at the
     * beginning of the series is patched up if necessary.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalStateException If the output has not been set, or if
     *                               prepareWriteBlobSequence has not been
     *                               called.
     */
    public void endWriteBlobSequence() throws IOException
    {
	// Reset the metadata
	_meta = null;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Blob Writer.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the output stream
	try { if (_output != null) { ((ImageOutputStream)_output).close(); } }
	catch (Exception e) {}

	_output = null;
    }

    /**
     * Gets the output stream.
     *
     * @return Image Output Stream to write to.
     *
     * @throws IllegalStateException If the output has not been set.
     */
    private ImageOutputStream _getOutputStream()
    {
	// No output has been set
	if (_output == null) {
	    String msg = "No output has been set.";
	    throw new IllegalStateException(msg);
	}

	// Return the Image Output Stream for the metadata
	return (ImageOutputStream)_output;
    }

    /**
     * Adds padding to the numbers.
     *
     * @param nums Numbers to pad.
     *
     * @return Padded numbers.
     */
    private static String[] _pad(String[] nums)
    {
	// Find the maximum length to the decimal
	int maxLength = 0;
	for (int i = 0; i < nums.length; i++) {
	    int length = nums[i].length();

	    // Decimal exists
	    int index = nums[i].indexOf('.');
	    if (index != -1) { length = index; }
	    
	    // Update the max
	    if (length > maxLength) { maxLength = length; }
	}

	// Pad the numbers so the decimals are aligned
	String[] paddedNums = new String[nums.length];
	for (int i = 0; i < nums.length; i++) {
	    int length = nums[i].length();

	    // Decimal exists
	    int index = nums[i].indexOf('.');
	    if (index != -1) { length = index; }

	    // Add padding
	    paddedNums[i] = "   " + nums[i];
	    for (int j = 0; j < (maxLength-length); j++) {
		paddedNums[i] = " " + paddedNums[i];
	    }
	}

	// Return the padded numbers
	return paddedNums;
    }
}
