/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.imageio.AppletFriendlyIIOMetadataFormat;
import javax.imageio.ImageTypeSpecifier;

/**
 * IIO Metadata Format that describes the tree structure of metadata for the
 * Free Surfer XFM Format.
 *
 * @version 14 January 2010
 */
public class XfmFormat extends AppletFriendlyIIOMetadataFormat
{
    /** Name of the native metadata format. */
    public static final String NATIVE_METADATA_FORMAT_NAME =
	"edu_ucla_loni_freesurfer_xfm_1.0";

    /** Transform type. */
    public static final String TRANSFORM_TYPE = "Transform_Type";

    /** Comment. */
    public static final String COMMENT = "Comment";

    /** Text of a comment. */
    public static final String COMMENT_TEXT = "Text";

    /** Linear transform. */
    public static final String LINEAR_TRANSFORM = "Linear_Transform";

    /** 1,1 Element of the linear transform. */
    public static final String ELEM_1_1 = "elem_1_1";

    /** 1,2 Element of the linear transform. */
    public static final String ELEM_1_2 = "elem_1_2";

    /** 1,3 Element of the linear transform. */
    public static final String ELEM_1_3 = "elem_1_3";

    /** 1,4 Element of the linear transform. */
    public static final String ELEM_1_4 = "elem_1_4";

    /** 2,1 Element of the linear transform. */
    public static final String ELEM_2_1 = "elem_2_1";

    /** 2,2 Element of the linear transform. */
    public static final String ELEM_2_2 = "elem_2_2";

    /** 2,3 Element of the linear transform. */
    public static final String ELEM_2_3 = "elem_2_3";

    /** 2,4 Element of the linear transform. */
    public static final String ELEM_2_4 = "elem_2_4";

    /** 3,1 Element of the linear transform. */
    public static final String ELEM_3_1 = "elem_3_1";

    /** 3,2 Element of the linear transform. */
    public static final String ELEM_3_2 = "elem_3_2";

    /** 3,3 Element of the linear transform. */
    public static final String ELEM_3_3 = "elem_3_3";

    /** 3,4 Element of the linear transform. */
    public static final String ELEM_3_4 = "elem_3_4";

    /** Single instance of the XFM Format. */
    private static XfmFormat _format = new XfmFormat();

    /** Constructs an XFM Format. */
    private XfmFormat()
    {
	// Create the root with optional children
	super(NATIVE_METADATA_FORMAT_NAME, CHILD_POLICY_SOME);

	// Add the transform type
	addAttribute(NATIVE_METADATA_FORMAT_NAME, TRANSFORM_TYPE,
		     DATATYPE_STRING, true, null);

	// Add the comment children
	addElement(COMMENT, NATIVE_METADATA_FORMAT_NAME, CHILD_POLICY_EMPTY);

	// Add the comment attributes
	addAttribute(COMMENT, COMMENT_TEXT, DATATYPE_STRING, true, null);

	// Add the linear transform child
	addElement(LINEAR_TRANSFORM, NATIVE_METADATA_FORMAT_NAME,
		   CHILD_POLICY_EMPTY);

	// Add the linear transform attributes
	addAttribute(LINEAR_TRANSFORM, ELEM_1_1, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_1_2, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_1_3, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_1_4, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_2_1, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_2_2, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_2_3, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_2_4, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_3_1, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_3_2, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_3_3, DATATYPE_STRING, true, null);
	addAttribute(LINEAR_TRANSFORM, ELEM_3_4, DATATYPE_STRING, true, null);
    }

    /**
     * Gets the instance of the XFM Format.
     *
     * @return Instance of the XFM Format.
     */
    public static XfmFormat getInstance()
    {
	return _format;
    }

    /**
     * Returns true if the element (and the subtree below it) is allowed to
     * appear in a metadata document for an image of the given type, defined by
     * an ImageTypeSpecifier.
     *
     * @param elementName The name of the element being queried.
     * @param imageType An ImageTypeSpecifier indicating the type of the image
     *                  that will be associated with the metadata.
     *
     * @return True if the node is meaningful for images of the given type.
     */
    public boolean canNodeAppear(String elementName,
				 ImageTypeSpecifier imageType)
    {
	// Not implemented
	return true;
    }
}
