/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.TreeMap;

/**
 * Registry of DICOM Data Elements.  For each Data Element, a Data Element
 * Tag, description, Value Representation, and Value Multiplicity is defined.
 * <p>
 * <b> PS 3.6-1998, DATA DICTIONARY </b>
 *
 * @version 8 April 2011
 */
public class DataDictionary
{
    /** Application entity value representation. */
    protected static String AE = "AE";

    /** Age string value representation. */
    protected static String AS = "AS";

    /** Attribute tag value representation. */
    protected static String AT = "AT";

    /** Code string value representation. */
    protected static String CS = "CS";

    /** Date value representation. */
    protected static String DA = "DA";

    /** Decimal string value representation. */
    protected static String DS = "DS";

    /** Date time value representation. */
    protected static String DT = "DT";

    /** Floating point double value representation. */
    protected static String FD = "FD";

    /** Floating point single value representation. */
    protected static String FL = "FL";

    /** Integer string value representation. */
    protected static String IS = "IS";

    /** Long string value representation. */
    protected static String LO = "LO";

    /** Long text value representation. */
    protected static String LT = "LT";

    /** Other byte string value representation. */
    protected static String OB = "OB";

    /** Other word string value representation. */
    protected static String OW = "OW";

    /** Person name value representation. */
    protected static String PN = "PN";

    /** Short string value representation. */
    protected static String SH = "SH";

    /** Signed long value representation. */
    protected static String SL = "SL";

    /** Sequence of items value representation. */
    protected static String SQ = "SQ";

    /** Signed short value representation. */
    protected static String SS = "SS";

    /** Short text value representation. */
    protected static String ST = "ST";

    /** Time value representation. */
    protected static String TM = "TM";

    /** Unique identifier value representation. */
    protected static String UI = "UI";

    /** Unsigned long value representation. */
    protected static String UL = "UL";

    /** Unsigned short value representation. */
    protected static String US = "US";

    /** Unlimited text value representation. */
    protected static String UT = "UT";

    /** Value multiplicity 1. */
    protected static String VM_1 = "1";

    /** Value multiplicity 1-n. */
    protected static String VM_1_N = "1-n";

    /** Single instance of the Data Dictionary. */
    private static DataDictionary _dictionary = new DataDictionary();

    /**
     * Data structure representing the registry of DICOM Data Elements.
     * <p>
     * <b> PS 3.6-1998, DATA DICTIONARY [Section 6], Page 5 </b>
     */
    private TreeMap _registry;

    /** Last accessed Data Element information from the registry. */
    private String[] _lastInfo;

    /** Value representation map. */
    private HashMap _vrs;

    /** Constructs a Data Dictionary. */
    protected DataDictionary()
    {
	// Create the registry
	_registry = new TreeMap();
	_lastInfo = null;

	// Store the value representations
	_vrs = new HashMap();
	_vrs.put(AE, AE);
	_vrs.put(AS, AS);
	_vrs.put(AT, AT);
	_vrs.put(CS, CS);
	_vrs.put(DA, DA);
	_vrs.put(DS, DS);
	_vrs.put(DT, DT);
	_vrs.put(FD, FD);
	_vrs.put(FL, FL);
	_vrs.put(IS, IS);
	_vrs.put(LO, LO);
	_vrs.put(LT, LT);
	_vrs.put(OB, OB);
	_vrs.put(OW, OW);
	_vrs.put(PN, PN);
	_vrs.put(SH, SH);
	_vrs.put(SL, SL);
	_vrs.put(SQ, SQ);
	_vrs.put(SS, SS);
	_vrs.put(ST, ST);
	_vrs.put(TM, TM);
	_vrs.put(UI, UI);
	_vrs.put(UL, UL);
	_vrs.put(US, US);
	_vrs.put(UT, UT);

	// Load the registry
	_loadRegistry();
    }

    /**
     * Gets the instance of the Data Dictionary.
     *
     * @return Single instance of the Data Dictionary.
     */
    public static DataDictionary getInstance()
    {
	return _dictionary;
    }

    /**
     * Gets the static value of the Value Representation.
     *
     * @param vr Value Representation.
     *
     * @return Static value of the Value Representation or the original value
     *         if no static value exists.
     */
    public String getStaticValueRepresentation(String vr)
    {
	String staticVr = (String)_vrs.get(vr);

	if (staticVr != null) { return staticVr; }
	return vr;
    }

    /**
     * Gets the Value Representation of the Data Element corresponding to
     * the Data Element Tag.
     *
     * @param dataElementTag Tag that uniquely identifies the Data Element.
     *                       This tag is represented as an 8 digit hexadecimal
     *                       number.  The first 4 digits represent the Group
     *                       Number and the last 4 digits represent the Element
     *                       Number.
     *
     * @return Two character code that specifies the data type and format of
     *         the Values of the Data Element, or null if it does not exist.
     */
    public String getValueRepresentation(String dataElementTag)
    {
	// Get the matching vector from the registry
	String[] values = _get(dataElementTag);
	if (values == null) { return null; }

	// VR is the first object in the Vector
	return values[0];
    }

    /**
     * Gets the name of the Data Element corresponding to the Data Element Tag.
     *
     * @param dataElementTag Tag that uniquely identifies the Data Element.
     *                       This tag is represented as an 8 digit hexadecimal
     *                       number.  The first 4 digits represent the Group
     *                       Number and the last 4 digits represent the Element
     *                       Number.
     *
     * @return Description of the Data Element, or null if it does not exist.
     */
    public String getName(String dataElementTag)
    {
	// Get the matching vector from the registry
	String[] values = _get(dataElementTag);
	if (values == null) { return null; }

	// Name is the second object in the Vector
	return values[1];
    }

    /**
     * Gets the Value Multiplicity of the Data Element corresponding to the
     * Data Element Tag.
     * <p>
     * <b> PS 3.5-1998, VALUE MULTIPLICITY (VM) AND DELIMITATION
     *     [Section 6.4], Page 23: </b>
     * <em>
     * The Value Multiplicity of a Data Element specifies the number of Values
     * that can be encoded in the Value Field of that Data Element. The VM of
     * each Data Element is specified explicitly in PS 3.6. If the number
     * of Values that may be encoded in an element is variable, it shall be
     * represented by two numbers separated by a dash; e.g., "1-10" means that
     * there may be 1 to 10 Values in the element.
     * </em>
     *
     * @param dataElementTag Tag that uniquely identifies the Data Element.
     *                       This tag is represented as an 8 digit hexadecimal
     *                       number.  The first 4 digits represent the Group
     *                       Number and the last 4 digits represent the Element
     *                       Number.
     *
     * @return Code which represents the number of Values that may be contained
     *         in the Data Element, or null if it does not exist.
     */
    public String getValueMultiplicity(String dataElementTag)
    {
	// Get the matching vector from the registry
	String[] values = _get(dataElementTag);
	if (values == null) { return null; }

	// VM is the third object in the Vector
	return values[2];
    }

    /**
     * Gets information about the Data Element.
     *
     * @param dataElementTag Tag that uniquely identifies the Data Element.
     *                       This tag is represented as an 8 digit hexadecimal
     *                       number.  The first 4 digits represent the Group
     *                       Number and the last 4 digits represent the Element
     *                       Number.
     *
     * @return Data Element Tag in the Data Dictionary, or null if it does not
     *         exist.
     */
    public String getDataElementTag(String dataElementTag)
    {
	// Get the matching vector from the registry
	String[] values = _get(dataElementTag);
	if (values == null) { return null; }

	// Tag is the fourth object in the Vector
	return values[3];
    }

    /**
     * Gets all the Data Element Tags defined by the Data Dictionary.
     *
     * @return All the Data Element Tags defined by the Data Dictionary listed
     *         in ascending order.  Each tag is represented as an 8 digit
     *         hexadecimal number.  The first 4 digits represent the Group
     *         Number and the last 4 digits represent the Element Number.
     */
    public String[] getDataElementTags()
    {
	// Get all the Data Element Tags as Objects
	Object[] keys = _registry.keySet().toArray();

	// Create a String array from the Object array
	String[] dataElementTags = new String[keys.length];
	System.arraycopy(keys, 0, dataElementTags, 0, keys.length);

	// Return the Data Element Tags
	return dataElementTags;
    }

    /** 
     * Gets the contents of the DataDictionary.
     *
     * @return String representation of the contents of the DataDictionary.
     */
    public String toString()
    {
	StringBuffer buffer = new StringBuffer();

	// Write the header
	buffer.append("TAG" + "\t" + "PROPERTIES" + "\n");

	// Write each tag and its corresponding values
	Iterator it = _registry.entrySet().iterator();
	while ( it.hasNext() ) {
	    Entry e = (Entry)it.next();

	    // Write the tag
	    buffer.append( e.getKey() );

	    // Write the Name
	    String[] values = (String[])e.getValue();
	    buffer.append("\t" + values[1] + "\n");

	    // Write the VR
	    buffer.append("\t" + "VR:  " + values[0] + "\n");

	    // Write the VM
	    buffer.append("\t" + "VM:  " + values[2] + "\n");
	}

	return buffer.toString();
    }

    /** Loads the registry of DICOM Data Elements. */
    protected void _loadRegistry()
    {
	// Copy the public dictionary
	DataDictionary publicDictionary = PublicDataDictionary.getInstance();
	_registry = (TreeMap)publicDictionary._registry.clone();

	// Add the private dictionaries
	_add( GeDataDictionary.getInstance() );
	_add( PhilipsDataDictionary.getInstance() );
	_add( SiemensDataDictionary.getInstance() );
    }

    /**
     * Adds an entry to the registry of DICOM Data Elements.
     *
     * @param dataElementTag Tag that uniquely identifies the Data Element.
     *                       This tag is represented as an 8 digit hexadecimal
     *                       number.  The first 4 digits represent the Group
     *                       Number and the last 4 digits represent the Element
     *                       Number.
     * @param valueRepresentation Two character code that specifies the data
     *                            type and format of the Values of the
     *                            Data Element.
     * @param name Description of the Data Element.
     * @param valueMultiplicity Code which represents the number of Values
     *                          that may be contained in the Data Element.
     */
    protected void _add(String dataElementTag, String valueRepresentation,
			String name, String valueMultiplicity)
    {
	// Pair the VR, name, VM, and tag
	String[] values = new String[4];
	values[0] = valueRepresentation;
	values[1] = name;
	values[2] = valueMultiplicity;
	values[3] = dataElementTag;

	// Add the pairing using the tag as the key
	_registry.put(dataElementTag, values);
    }

  /**
   * Gets an entry from the registry of DICOM Data Elements.
   *
   * @param dataElementTag Tag that uniquely identifies the Data Element.
   *                       This tag is represented as an 8 digit hexadecimal
   *                       number.  The first 4 digits represent the Group
   *                       Number and the last 4 digits represent the Element
   *                       Number.
   *
   * @return Object array containing the Value Representation, name, and
   *         Value Multiplicity corresponding to the Data Element Tag, or null
   *         if this array does not exist.
   */
    private synchronized String[] _get(String dataElementTag)
    {
	String tag = dataElementTag;

	// Check the cache
	if (_lastInfo != null && _lastInfo[3].equals(tag)) { return _lastInfo; }

	// If the tag is from group 5000 or group 6000, replace the last two
	// hexadecimal digits of the Group Number with "xx"
	if ( tag.startsWith("50") || tag.startsWith("60") ) {
	    StringBuffer tagBuffer = new StringBuffer(tag);
	    tagBuffer.replace(2, 4, "xx");
	    tag = tagBuffer.toString();
	}

	// Check the registry for a match on the tag
	Object o = _registry.get(tag);
	if (o == null) {
	    _lastInfo = null;
	    return null;
	}

	_lastInfo = (String [])o;
	return _lastInfo;
    }

    /**
     * Adds the Data Dictionary to this Data Dictionary.  If there are duplicate
     * Data Element Tags, those tags are removed from this Data Dictionary.
     *
     * @param dictionary Data Dictionary to add.
     */
    private void _add(DataDictionary dictionary)
    {
	// Add the tags
	Iterator iter = dictionary._registry.entrySet().iterator();
	while ( iter.hasNext() ) {
	    Entry e = (Entry)iter.next();

	    // Tag exists in this dictionary
	    if ( _registry.containsKey(e.getKey()) ) {

		// Remove the tag from this dictionary
		_registry.remove(e.getKey());
	    }

	    // Tag does not exist in this dictionary
	    else {

		// Add the tag to this dictionary
		_registry.put(e.getKey(), e.getValue());
	    }
	}
    }
}
