/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.node;

import java.util.ArrayList;
import java.util.Iterator;
import org.w3c.dom.DOMException;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * DICOM Node that has children.
 *
 * @version 5 April 2010
 */
public abstract class DicomListNode extends DicomNode implements NodeList
{
    /** Child Nodes. */
    private ArrayList _children;

    /**
     * Constructs a DICOM List Node.
     *
     * @param children Child Nodes.
     */
    protected DicomListNode(ArrayList children)
    {
	_children = children;

	// Link the children together
	DicomNode prevNode = null;
	Iterator iter = _children.iterator();
	while ( iter.hasNext() ) {
	    DicomNode childNode = (DicomNode)iter.next();

	    // Set the parent
	    childNode._setParent(this);

	    // Set the siblings
	    childNode._setPreviousSibling(prevNode);
	    if (prevNode != null) { prevNode._setNextSibling(childNode); }
	    prevNode = childNode;
	}
    }

    /**
     * Gets an iterator over the child Nodes.
     *
     * @return Iterator over the child Nodes.
     */
    public Iterator getChildNodeIterator()
    {
	return new DicomNodeIterator( _children.iterator() );
    }

    /**
     * Gets the Node type.
     *
     * @return Node type.
     */
    public short getNodeType()
    {
	return ELEMENT_NODE;
    }

    /**
     * Gets a Node List that contains all children of this Node.
     *
     * @return Node List that contains all children of this Node.
     */
    public NodeList getChildNodes()
    {
	return this;
    }

    /**
     * Gets the first child of this Node.
     *
     * @return First child of this Node or null if none exists.
     */
    public Node getFirstChild()
    {
	return item(0);
    }

    /**
     * Gets the last child of this Node.
     *
     * @return Last child of this Node or null if none exists.
     */
    public Node getLastChild()
    {
	return item(_children.size()-1);
    }

    /**
     * Inserts the Node before the existing child Node.  If the existing child
     * Node is null, the new Node is inserted at the end of the list of
     * children.
     *
     * @param newChild Node to insert.
     * @param refChild Node before which the new Node is inserted.
     *
     * @return Node that is inserted.

     * @throws DOMException If a child is not a DICOM Node.
     * @throws DOMException If the reference child is not a child of this Node.
     */
    public Node insertBefore(Node newChild, Node refChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof DicomNode) ) {
	    String msg = "Illegal new child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DicomNode newNode = (DicomNode)newChild;

	// No reference child
	if (refChild == null) { return appendChild(newNode); }

	// Illegal reference child
	if ( !(refChild instanceof DicomNode) ) {
	    String msg = "Illegal reference child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DicomNode refNode = (DicomNode)refChild;

	// Find the reference child
	for (int i = 0; i < _children.size(); i++) {
	    if (refNode == _children.get(i)) {

		// Insert the new child before the reference child
		_children.add(i, newNode);

		// Set the parent
		newNode._setParent(this);

		// Reset the siblings
		DicomNode prevNode = (DicomNode)refNode.getPreviousSibling();
		newNode._setPreviousSibling(prevNode);
		if (prevNode != null) { prevNode._setNextSibling(newNode); }
		newNode._setNextSibling(refNode);
		refNode._setPreviousSibling(newNode);

		// Return the Node that is inserted
		return newNode;
	    }
	}

	// Cannot find the reference child
	String msg = "Reference child does not exist";
	throw new DOMException(DOMException.NOT_FOUND_ERR, msg);
    }

    /**
     * Replaces the child Node with the given child Node.
     *
     * @param newChild Node to insert.
     * @param oldChild Node being replaced.
     *
     * @return Node that is replaced.
     *
     * @throws DOMException If a child is not a DICOM Node.
     * @throws DOMException If the old child is not a child of this Node.
     */
    public Node replaceChild(Node newChild, Node oldChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof DicomNode) ) {
	    String msg = "Illegal new child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DicomNode newNode = (DicomNode)newChild;

	// Illegal current child
	if ( !(oldChild instanceof DicomNode) ) {
	    String msg = "Illegal old child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DicomNode oldNode = (DicomNode)oldChild;

	// Find the current child
	for (int i = 0; i < _children.size(); i++) {
	    if (oldNode == _children.get(i)) {

		// Replace the current child with the new child
		_children.set(i, newNode);

		// Set the parent
		newNode._setParent(this);

		// Transfer the siblings
		newNode._setPreviousSibling( oldNode.getPreviousSibling() );
		newNode._setNextSibling( oldNode.getNextSibling() );

		// Erase references of the old child
		oldNode._setParent(null);
		oldNode._setPreviousSibling(null);
		oldNode._setNextSibling(null);

		// Return the Node that was replaced
		return oldNode;
	    }
	}

	// Cannot find the current child
	String msg = "Current child does not exist";
	throw new DOMException(DOMException.NOT_FOUND_ERR, msg);
    }

    /**
     * Removes the child Node from the list of children Nodes.
     *
     * @param oldChild Node being removed.
     *
     * @return Node that is removed.
     *
     * @throws DOMException If the child is not a DICOM Node.
     * @throws DOMException If the old child is not a child of this Node.
     */
    public Node removeChild(Node oldChild) throws DOMException
    {
	// Illegal current child
	if ( !(oldChild instanceof DicomNode) ) {
	    String msg = "Illegal old child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DicomNode oldNode = (DicomNode)oldChild;

	// Find the child
	for (int i = 0; i < _children.size(); i++) {
	    if (oldNode == _children.get(i)) {

		// Remove the current child
		_children.remove(i);

		// Reconnect the siblings
		DicomNode prevNode = (DicomNode)oldNode.getPreviousSibling();
		DicomNode nextNode = (DicomNode)oldNode.getNextSibling();
		if (prevNode != null) { prevNode._setNextSibling(nextNode); }
		if (nextNode != null) {nextNode._setPreviousSibling(prevNode);}

		// Erase references of the old child
		oldNode._setParent(null);
		oldNode._setPreviousSibling(null);
		oldNode._setNextSibling(null);

		// Return the Node that was removed
		return oldNode;
	    }
	}

	// Cannot find the current child
	String msg = "Child does not exist";
	throw new DOMException(DOMException.NOT_FOUND_ERR, msg);
    }

    /** Removes all the children Nodes. */
    public void removeAllChildren()
    {
	// Erase references of each child
	for (int i = 0; i < _children.size(); i++) {
	    DicomNode childNode = (DicomNode)_children.get(i);

	    childNode._setParent(null);
	    childNode._setPreviousSibling(null);
	    childNode._setNextSibling(null);
	}

	// Remove the children
	_children = new ArrayList();
    }

    /**
     * Adds the Node to the end of the list of children Nodes.
     *
     * @param newChild Node to add.
     *
     * @return Node that is added.
     *
     * @throws DOMException If the child is not a DICOM Node.
     */
    public Node appendChild(Node newChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof DicomNode) ) {
	    String msg = "Illegal new child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DicomNode newNode = (DicomNode)newChild;

	// Add the new child
	_children.add(newNode);

	// Set the parent
	newNode._setParent(this);

	// Connect the siblings
	int prevIndex = _children.size()-2;
	if (prevIndex >= 0) {
	    DicomNode prevNode = (DicomNode)_children.get(prevIndex);
	    prevNode._setNextSibling(newNode);
	    newNode._setPreviousSibling(prevNode);
	}

	// Return the Node that was added
	return newNode;
    }

    /**
     * Determines if this Node has child Nodes.
     *
     * @return True if this Node has child Nodes; false otherwise.
     */
    public boolean hasChildNodes()
    {
	return true;
    }

    /**
     * Gets the indexed child Node.
     *
     * @param index Index of the child.
     *
     * @return Indexed child Node or null if the index is invalid.
     */
    public Node item(int index)
    {
	try { return (Node)_children.get(index); }
	catch (IndexOutOfBoundsException e) { return null; }
    }

    /**
     * Gets the number of children of this Node.
     *
     * @return Number of children of this Node.
     */
    public int getLength()
    {
	return _children.size();
    }
}
