/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.node;

import edu.ucla.loni.jdicom.DataSet;
import org.w3c.dom.DOMException;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * DICOM Node for the root of a DICOM tree.
 *
 * @version 2 April 2010
 */
public class DicomRootNode extends DicomNode implements NodeList
{
    /** Root name. */
    private String _rootName;

    /** Data Set Node. */
    private DataSetNode _dataSet;

    /**
     * Constructs a DICOM Root Node.
     *
     * @param rootName Root name.
     * @param dataSet Data Set to convert.
     */
    public DicomRootNode(String rootName, DataSet dataSet)
    {
	this(rootName, new DataSetNode(dataSet));
    }

    /**
     * Constructs a DICOM Root Node.
     *
     * @param rootName Root name.
     * @param dataSet Data Set Node.
     */
    private DicomRootNode(String rootName, DataSetNode dataSet)
    {
	_rootName = rootName;
	_dataSet = dataSet;
    }

    /**
     * Gets the Data Set Node.
     *
     * @return Data Set Node.
     */
    public DataSetNode getDataSet()
    {
	return _dataSet;
    }

    /**
     * Gets the Node name.
     *
     * @return Name of the Node.
     */
    public String getNodeName()
    {
	return _rootName;
    }

    /**
     * Gets the Node type.
     *
     * @return Node type.
     */
    public short getNodeType()
    {
	return ELEMENT_NODE;
    }

    /**
     * Gets a Node List that contains all children of this Node.
     *
     * @return Node List that contains all children of this Node.
     */
    public NodeList getChildNodes()
    {
	return this;
    }

    /**
     * Gets the first child of this Node.
     *
     * @return First child of this Node or null if none exists.
     */
    public Node getFirstChild()
    {
	return _dataSet;
    }

    /**
     * Gets the last child of this Node.
     *
     * @return Last child of this Node or null if none exists.
     */
    public Node getLastChild()
    {
	return _dataSet;
    }

    /**
     * Inserts the Node before the existing child Node.  If the existing child
     * Node is null, the new Node is inserted at the end of the list of
     * children.
     *
     * @param newChild Node to insert.
     * @param refChild Node before which the new Node is inserted.
     *
     * @return Node that is inserted.

     * @throws DOMException Since this is a read-only Node.
     */
    public Node insertBefore(Node newChild, Node refChild) throws DOMException
    {
	String msg = "Node is read-only";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Replaces the child Node with the given child Node.
     *
     * @param newChild Node to insert.
     * @param oldChild Node being replaced.
     *
     * @return Node that is replaced.

     * @throws DOMException Since this is a read-only Node.
     */
    public Node replaceChild(Node newChild, Node oldChild) throws DOMException
    {
	String msg = "Node is read-only";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Removes the child Node from the list of children Nodes.
     *
     * @param oldChild Node being removed.
     *
     * @return Node that is removed.

     * @throws DOMException Since this is a read-only Node.
     */
    public Node removeChild(Node oldChild) throws DOMException
    {
	String msg = "Node is read-only";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Adds the Node to the end of the list of children Nodes.
     *
     * @param newChild Node to add.
     *
     * @return Node that is added.

     * @throws DOMException Since this is a read-only Node.
     */
    public Node appendChild(Node newChild) throws DOMException
    {
	String msg = "Node is read-only";
	throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, msg);
    }

    /**
     * Determines if this Node has child Nodes.
     *
     * @return True if this Node has child Nodes; false otherwise.
     */
    public boolean hasChildNodes()
    {
	return true;
    }

    /**
     * Gets a duplicate of this Node.  The duplicate Node has no parent.  If a
     * shallow clone is being performed, the duplicate Node will not have any
     * children or siblings.  If a deep clone is performed, the duplicate Node
     * will form the root of a completely cloned subtree.
     *
     * @param deep True for a recursive clone, false for cloning only the Node
     *             itself.
     *
     * @return Duplicate Node.
     */
    public Node cloneNode(boolean deep)
    {
	DataSetNode dataSet = (DataSetNode)_dataSet.cloneNode(deep);
	return new DicomRootNode(_rootName, dataSet);
    }

    /**
     * Gets the indexed child Node.
     *
     * @param index Index of the child.
     *
     * @return Indexed child Node or null if the index is invalid.
     */
    public Node item(int index)
    {
	if (index == 0) { return _dataSet; }
	return null;
    }

    /**
     * Gets the number of children of this Node.
     *
     * @return Number of children of this Node.
     */
    public int getLength()
    {
	return 1;
    }
}
