/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.write;

import edu.ucla.loni.jdicom.DataSet;
import java.awt.image.DataBuffer;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;
import javax.imageio.stream.ImageOutputStream;

/**
 * Writes DICOM information to an output stream.  The stream is structured as:
 *
 * <p>
 * ++++++++++++++++++++++++++++++++++++++++++++++++++ --
 * + Preamble and "DICM"                            +
 * ++++++++++++++++++++++++++++++++++++++++++++++++++ Sequence Header
 * + Data Elements Before (5200,9230) Element       +
 * ++++++++++++++++++++++++++++++++++++++++++++++++++ --
 * + Data Sets in (5200,9230) Element               + Image Metadata
 * ++++++++++++++++++++++++++++++++++++++++++++++++++ --
 * + (7FE0,0010) Element Tag, VR, and Value Length  +
 * ++++++++++++++++++++++++++++++++++++++++++++++++++
 * + Pixel Data for Image 1                         +
 * ++++++++++++++++++++++++++++++++++++++++++++++++++
 * + Pixel Data for Image 2                         + Image Sequence
 * ++++++++++++++++++++++++++++++++++++++++++++++++++
 * + ...                                            +
 * ++++++++++++++++++++++++++++++++++++++++++++++++++
 * + Pixel Data for Image N                         +
 * ++++++++++++++++++++++++++++++++++++++++++++++++++ --
 * + Data Elements After (7FE0, 0010) Element       + Sequence Footer
 * ++++++++++++++++++++++++++++++++++++++++++++++++++ --
 *
 * @version 16 March 2011
 */
public class DicomWriter
{
    /** Output stream to write DICOM information to. */
    private ImageOutputStream _outputStream;

    /** Transfer syntax that determines the encoding. */
    private String _transferSyntax;

    /** Image metadata block. */
    private ImageMetadata _imageMetadata;

    /** Image sequence block. */
    private ImageSequence _imageSequence;

    /** Cached image metadata data sets. */
    private Vector _imageMetadataDataSets;

    /**
     * Constructs a DICOM Writer.
     *
     * @param outputStream Output stream to write DICOM information to.
     * @param transferSyntax Transfer syntax that determines the encoding.
     */
    public DicomWriter(ImageOutputStream outputStream, String transferSyntax)
    {
	_outputStream = outputStream;
	_transferSyntax = transferSyntax;
	_imageMetadataDataSets = new Vector();
    }

    /**
     * Writes the sequence header.
     *
     * @param dataSet Data Set containing the Data Elements to write.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void writeSequenceHeader(DataSet dataSet) throws IOException
    {
	// Write the sequence header
	new SequenceHeader(_outputStream, _transferSyntax, dataSet);
    }

    /**
     * Writes metadata for an image of the image sequence.  If this method is
     * called then all additional image metadata must be written using this
     * method.  All metadata must be written before writing image pixel data
     * using writeImage(DataBuffer).
     *
     * @param dataSet Data Set containing the image metadata elements to write.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void writeImageMetadata(DataSet dataSet) throws IOException
    {
	// Initialize the image metadata block
	if (_imageMetadata == null) {
	    _imageMetadata = new ImageMetadata(_outputStream, _transferSyntax);
	}

	// Replace missing data sets with an empty data set
	if (dataSet == null) { dataSet = new DataSet(); }

	// Write the image metadata
	_imageMetadata.append(dataSet);
    }

    /**
     * Writes an image of the image sequence.  If this method is called then
     * all additional images must be written using this method.
     *
     * @param dataBuffer Data Buffer containing the pixel data.
     * @param width Image width.
     * @param height Image height.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void writeImage(DataBuffer dataBuffer, int width, int height)
	throws IOException
    {
	// Finalize the image metadata block
	if (_imageMetadata != null) {
	    _imageMetadata.finish();
	    _imageMetadata = null;
	}

	// Initialize the image sequence
	if (_imageSequence == null) {
	    _imageSequence = new ImageSequence(_outputStream, _transferSyntax);
	}

	// Write the image
	_imageSequence.append(dataBuffer, width, height);
    }

    /**
     * Writes an image of the image sequence.  If this method is called then
     * all additional images and their metadata must be written using this
     * method.
     *
     * @param dataSet Data Set containing the image metadata elements to write.
     * @param dataBuffer Data Buffer containing the pixel data.
     * @param width Image width.
     * @param height Image height.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void writeImage(DataSet dataSet, DataBuffer dataBuffer, int width,
			   int height)
	throws IOException
    {
	// Replace missing data sets with an empty data set
	if (dataSet == null) { dataSet = new DataSet(); }

	// Cache the image metadata
	_imageMetadataDataSets.add(dataSet);

	// Initialize the image sequence
	if (_imageSequence == null) {
	    _imageSequence = new ImageSequence(_outputStream, _transferSyntax);
	}

	// Write the image
	_imageSequence.append(dataBuffer, width, height);
    }

    /**
     * Writes the sequence footer.
     *
     * @param dataSet Data Set containing the Data Elements to write.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void writeSequenceFooter(DataSet dataSet) throws IOException
    {
	// Finalize the image sequence
	if (_imageMetadata != null) { _imageMetadata.finish(); }
	if (_imageSequence != null) { _imageSequence.finish(); }

	// Write the sequence footer
	SequenceFooter footer = new SequenceFooter(_outputStream,
						   _transferSyntax, dataSet);

	// Cached image metadata to write
	if (_imageMetadataDataSets.size() > 0) {

	    // Offset where the image metadata will be written to
	    long imageMetadataOffset = _imageSequence.getOffset();

	    // Determine the size of the image metadata
	    CounterOutputStream cos = new CounterOutputStream();
	    ImageMetadataOS imgOs = new ImageMetadataOS(cos, _transferSyntax);
	    Iterator iter = _imageMetadataDataSets.iterator();
	    while ( iter.hasNext() ) {
		DataSet imgDataSet = (DataSet)iter.next();
		imgOs.append(imgDataSet);
	    }
	    imgOs.finish();

	    // Relocate the sequence footer and then the image sequence
	    long shift = cos.getNumberOfBytesWritten();
	    footer.moveTo(footer.getOffset() + shift);
	    _imageSequence.moveTo(_imageSequence.getOffset() + shift);

	    // Return to where the image metadata is to be written
	    _outputStream.seek(imageMetadataOffset);

	    // Write the image metadata
	    ImageMetadata imageMetadata = new ImageMetadata(_outputStream,
							    _transferSyntax);
	    iter = _imageMetadataDataSets.iterator();
	    while ( iter.hasNext() ) {
		DataSet imgDataSet = (DataSet)iter.next();
		imageMetadata.append(imgDataSet);
	    }
	    imageMetadata.finish();
	}
    }
}
