/*
  COPYRIGHT NOTICE
  Copyright (c) 20011  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nifti;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.ByteOrder;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.MemoryCacheImageInputStream;

/**
 * Decoder for the header of the NIFTI 1 format.
 *
 * @version 7 April 2011
 */
public class NiftiHeaderDecoder
{
    /** Decoded header. */
    private NiftiHeader _header;

    /** True if the encoded NIFTI Header is byte swapped; false otherwise. */
    private boolean _isByteSwapped;

    /**
     * Constructs a NIFTI Header Decoder.
     *
     * @param imageStream Input stream to the header.
     *
     * @throws IOException If an I/O error occurs.
     */
    public NiftiHeaderDecoder(ImageInputStream imageStream) throws IOException
    {
	// Read the encoded NIFTI header into a byte array and reset Stream
	byte[] headerBytes = new byte[348];
	imageStream.mark();
	imageStream.readFully(headerBytes);
	imageStream.reset();

	// Create a new Image Input Stream that reads from the bytes
	ByteArrayInputStream bStream = new ByteArrayInputStream(headerBytes);
	MemoryCacheImageInputStream mStream =
	    new MemoryCacheImageInputStream(bStream);

	// Read the number of dimensions
	mStream.mark();
	mStream.skipBytes(40);
	short dim0 = mStream.readShort();
	mStream.reset();

	// Determine if bytes should be swapped
	_isByteSwapped = false;
	if (dim0 < 0 || dim0 > 7) {
	    _isByteSwapped = true;
	    mStream.setByteOrder(ByteOrder.LITTLE_ENDIAN);
	}

	// Define the NIFTI Header field arrays
	byte[] dataType = new byte[10];
	byte[] dbName = new byte[18];
	short[] dim = new short[8];
	float[] intentP = new float[3];
	float[] pixDim = new float[8];
	byte[] descrip = new byte[80];
	byte[] auxFile = new byte[24];
	float[] quatern = new float[3];
	float[] qoffset = new float[3];
	float[] srow = new float[12];
	byte[] intentName = new byte[16];
	byte[] magic = new byte[4];

	// Read the NIFTI Header fields
	int sizeOfHdr = mStream.readInt();
	mStream.readFully(dataType);
	mStream.readFully(dbName);
	int extents = mStream.readInt();
	short sessionError = mStream.readShort();
	byte regular = mStream.readByte();
	byte dimInfo = mStream.readByte();
	mStream.readFully(dim, 0, dim.length);
	mStream.readFully(intentP, 0, intentP.length);
	short intentCode = mStream.readShort();
	short datatype = mStream.readShort();
	short bitPix = mStream.readShort();
	short sliceStart = mStream.readShort();
	mStream.readFully(pixDim, 0, pixDim.length);
	float voxOffset = mStream.readFloat();
	float sclSlope = mStream.readFloat();
	float sclInter = mStream.readFloat();
	short sliceEnd = mStream.readShort();
	byte sliceCode = mStream.readByte();
	byte xyztUnits = mStream.readByte();
	float calMax = mStream.readFloat();
	float calMin = mStream.readFloat();
	float sliceDuration = mStream.readFloat();
	float tOffset = mStream.readFloat();
	int glMax = mStream.readInt();
	int glMin = mStream.readInt();
	mStream.readFully(descrip);
	mStream.readFully(auxFile);
	short qformCode = mStream.readShort();
	short sformCode = mStream.readShort();
	mStream.readFully(quatern, 0, quatern.length);
	mStream.readFully(qoffset, 0, qoffset.length);
	mStream.readFully(srow, 0, srow.length);
	mStream.readFully(intentName);
	mStream.readFully(magic);

	// Construct the NIFTI Header
	_header =
	    new NiftiHeader(sizeOfHdr, dataType, dbName, extents,
			    sessionError, regular, dimInfo, dim[0], dim[1],
			    dim[2], dim[3], dim[4], dim[5], dim[6], dim[7],
			    intentP[0], intentP[1], intentP[2], intentCode,
			    datatype, bitPix, sliceStart, pixDim[0],
			    pixDim[1], pixDim[2], pixDim[3], pixDim[4],
			    pixDim[5], pixDim[6], pixDim[7], voxOffset,
			    sclSlope, sclInter, sliceEnd, sliceCode,
			    xyztUnits, calMax, calMin, sliceDuration,
			    tOffset, glMax, glMin, descrip, auxFile,
			    qformCode, sformCode, quatern[0], quatern[1],
			    quatern[2], qoffset[0], qoffset[1], qoffset[2],
			    srow[0], srow[1], srow[2], srow[3], srow[4],
			    srow[5], srow[6], srow[7], srow[8], srow[9],
			    srow[10], srow[11], intentName, magic);
    }

    /**
     * Gets the decoded header.
     *
     * @return Decoded header.
     */
    public NiftiHeader getHeader()
    {
	return _header;
    }

    /**
     * Determines if the header is byte swapped.
     *
     * @return True if the encoded ANALYZE Header is byte swapped; false o/w.
     */
    public boolean isByteSwapped()
    {
	return _isByteSwapped;
    }
}
