/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nifti;

import java.io.DataOutput;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import javax.imageio.stream.ImageOutputStream;

/**
 * Encoder for the header of the NIFTI 1 format.
 *
 * @version 7 April 2011
 */
public class NiftiHeaderEncoder
{
    /** Output stream for the header. */
    private DataOutput _imageStream;

    /** True if the output consists of two Image Input Streams; false if one. */
    private boolean _hasTwoStreams;

    /**
     * Constructs a Nifti Header Encoder.
     *
     * @param imageStream Output stream for the header.
     * @param hasTwoStreams True if the output consists of two Image Input
     *                      Streams; false if one.
     */
    public NiftiHeaderEncoder(DataOutput imageStream, boolean hasTwoStreams) 
    {
	_imageStream = imageStream;
	_hasTwoStreams = hasTwoStreams;
    }

    /**
     * Writes the header.
     *
     * @param header Header to encode.
     *
     * @return Number of Header bytes written.
     *
     * @throws IOException If an I/O error occurs.
     */
    public int writeHeader(NiftiHeader header) throws IOException
    {
	int numberOfBytesWritten = 0;

	Iterator iter = header.getFieldNames();
	while ( iter.hasNext() ) {
	    String fieldName = (String)iter.next();
	    Object fieldValue = header.getFieldValue(fieldName);

	    // Field value is a String
	    if (fieldValue instanceof String) {
		int size = 0;
		if ( NiftiHeader.DATA_TYPE.equals(fieldName) ) { size = 10; }
		if ( NiftiHeader.DB_NAME.equals(fieldName) ) { size = 18; }
		if ( NiftiHeader.DESCRIP.equals(fieldName) ) { size = 80; }
		if ( NiftiHeader.AUX_FILE.equals(fieldName) ) { size = 24; }
		if ( NiftiHeader.INTENT_NAME.equals(fieldName) ) { size = 16; }
		if ( NiftiHeader.MAGIC.equals(fieldName) ) { size = 4; }

		// Overwrite any value for the magic number
		if (NiftiHeader.MAGIC.equals(fieldName) ) {
		    if (_hasTwoStreams) { fieldValue = "ni1"; }
		    else { fieldValue = "n+1"; }
		}

		// Write the string value
		numberOfBytesWritten += _writeString((String)fieldValue, size);
	    }

	    // Field value is a Number
	    else { numberOfBytesWritten += _writeNumber( (Number)fieldValue ); }
	}

	// Write a blank extension if one stream
	if (!_hasTwoStreams) {
	    numberOfBytesWritten += _writeNumber( new Integer(0) );
	}

	// Return the number of bytes written
	return numberOfBytesWritten;
    }

    /**
     * Writes the Number to the output stream.
     *
     * @param number Number to write to the output stream.
     *
     * @return Number of Header bytes written.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private int _writeNumber(Number number) throws IOException
    {
	// Byte
	if (number instanceof Byte) {
	    int value = ((Byte)number).byteValue() & 0xffffffff;
	    _imageStream.writeByte(value);
	    return 1;
	}

	// Short
	else if (number instanceof Short) {
	    int value = ((Short)number).shortValue() & 0xffffffff;
	    _imageStream.writeShort(value);
	    return 2;
	}

	// Integer
	else if (number instanceof Integer) {
	    _imageStream.writeInt( ((Integer)number).intValue() );
	    return 4;
	}

	// Float
	else if (number instanceof Float) {
	    _imageStream.writeFloat( ((Float)number).floatValue() );
	    return 4;
	}

	// Unknown Number
	else {
	    String msg = "Unable to write an Number of type \"" +
		number.getClass().getName() + "\".";
	    throw new IOException(msg);
	}
    }

    /**
     * Writes the String to the output stream.
     *
     * @param string String to write to the output stream.
     * @param length Length of the string to actually write.
     *
     * @return Number of Header bytes written.
     *
     * @throws IOException If an error occurs during the writing.
     */
    private int _writeString(String string, int length) throws IOException
    {
	// Create a byte buffer of the specified length
	byte[] buffer = new byte[length];

	// Copy the bytes of the String to the byte buffer
	byte[] stringBytes = string.getBytes();
	int min = Math.min( buffer.length, stringBytes.length );
	System.arraycopy(stringBytes, 0, buffer, 0, min);

	// Write the bytes
	_imageStream.write(buffer);
	return length;
    }
}
