/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd;

import javax.imageio.IIOException;
import javax.imageio.stream.ImageOutputStream;

/**
 * Class that encodes an NRRD header.
 *
 * @version 29 October 2010
 */
public class NrrdHeaderEncoder
{
    /** Constructs an NRRD Header Encoder. */
    public NrrdHeaderEncoder()
    {
    }

    /**
     * Encodes the NRRD Header to the output stream.
     *
     * @param header NRRD Header to encode.
     * @param outputStream Image Output Stream to write to.
     * @param addEmptyLine True if an empty line is to be added at the end of
     *                     the encoding.
     *
     * @throws IIOException If an I/O error occurs.
     */
    public void encode(NrrdHeader header, ImageOutputStream outputStream,
		       boolean addEmptyLine)
	throws IIOException
    {
	try {

	    // Write the magic number
	    outputStream.writeBytes(header.getMagicNumber() + "\r\n");

	    // Write the header elements
	    NrrdHeaderElement[] elements = header.getElements();
	    for (int i = 0; i < elements.length; i++) {

		// Write the Field
		if (elements[i] instanceof Field) {
		    Field field = (Field)elements[i];

		    outputStream.writeBytes( field.getIdentifier() + ": " +
					     field.getDescriptor() + "\r\n");
		}

		// Write the Key-Value Pair
		else if (elements[i] instanceof KeyValuePair) {
		    KeyValuePair pair = (KeyValuePair)elements[i];

		    outputStream.writeBytes( pair.getKey() + ":=" +
					     pair.getValue() + "\r\n");
		}

		// Write the Key-Value Pair Set
		else if (elements[i] instanceof KeyValuePairSet) {
		    KeyValuePairSet set = (KeyValuePairSet)elements[i];
		    String name = set.getName();

		    // Write each Key-Value Pair
		    KeyValuePair[] pairs = set.getPairs();
		    for (int j = 0; j < pairs.length; j++) {

			outputStream.writeBytes(name + "_" + pairs[j].getKey() +
						":=" + pairs[j].getValue() +
						"\r\n");
		    }
		}
	    }

	    // Write a blank line
	    if (addEmptyLine) { outputStream.writeBytes("\r\n"); }
	}

	catch (Exception e) {
	    String msg = "Unable to encode the NRRD Header.";
	    throw new IIOException(msg, e);
	}
    }
}


