/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import edu.ucla.loni.imageio.Utilities;
import edu.ucla.loni.nrrd.NrrdHeader;
import edu.ucla.loni.nrrd.NrrdHeaderDecoder;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageReader;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageReaderSpi;
import javax.imageio.stream.ImageInputStream;

/**
 * Image Reader for parsing and decoding NRRD images.
 *
 * @version 16 March 2012
 */
public class NrrdImageReader extends ImageReader
{
    /** NRRD Header decoded from the input source. */
    private NrrdHeader _header;

    /** Pixel data image input stream. */
    private ImageInputStream _pixelDataStream;

    /**
     * Constructs an NRRD Image Reader.
     *
     * @param originatingProvider The Image Reader SPI that instantiated this
     *                            Object.
     *
     * @throws IllegalArgumentException If the originating provider is not an
     *                                  NRRD Image Reader Spi.
     */
    public NrrdImageReader(ImageReaderSpi originatingProvider)
    {
	super(originatingProvider);

	// Originating provider must be an NRRD Image Reader Spi
	if ( !(originatingProvider instanceof NrrdImageReaderSpi) ) {
	    String msg = "Originating provider must be an NRRD Image Reader " +
		"SPI.";
	    throw new IllegalArgumentException(msg);
	}
    }

    /**
     * Sets the input source to use to the given ImageInputStream or other
     * Object.  The input source must be set before any of the query or read
     * methods are used.  If the input is null, any currently set input source
     * will be removed.
     *
     * @param input The ImageInputStream or other Object to use for future
     *              decoding.
     * @param seekForwardOnly If true, images and metadata may only be read in
     *                        ascending order from this input source.
     * @param ignoreMetadata If true, metadata may be ignored during reads.
     *
     * @throws IllegalArgumentException If the input is not an instance of one
     *                                  of the classes returned by the
     *                                  originating service provider's
     *                                  getInputTypes method, or is not an
     *                                  ImageInputStream.
     * @throws IllegalStateException If the input source is an Input Stream that
     *                               doesn't support marking and seekForwardOnly
     *                               is false.
     */
    public void setInput(Object input, boolean seekForwardOnly,
			 boolean ignoreMetadata)
    {
	// Convert the input source
	input = NrrdImageReaderSpi.convertToInput(input);

	super.setInput(input, seekForwardOnly, ignoreMetadata);

	// Remove any decoded header
	_header = null;
	_pixelDataStream = null;
    }

    /**
     * Returns the number of images, not including thumbnails, available from
     * the current input source.
     * 
     * @param allowSearch If true, the true number of images will be returned
     *                    even if a search is required.  If false, the reader
     *                    may return -1 without performing the search.
     *
     * @return The number of images or -1 if allowSearch is false and a search
     *         would be required.
     *
     * @throws IllegalStateException If the input source has not been set.
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     */
    public int getNumImages(boolean allowSearch) throws IOException
    {
	NrrdMetadata metadata = new NrrdMetadata( _getHeader() );
	return metadata.getNumberOfImages();
    }

    /**
     * Returns the width in pixels of the given image within the input source.
     *
     * @param imageIndex The index of the image to be queried.
     *
     * @return The width of the image in pixels.
     *
     * @throws IOException If an error occurs reading the width information from
     *                     the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public int getWidth(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	_checkImageIndex(imageIndex);

	// Return the image width
	NrrdMetadata metadata = new NrrdMetadata( _getHeader() );
	return metadata.getImageWidth();
    }

    /**
     * Returns the height in pixels of the given image within the input source.
     *
     * @param imageIndex The index of the image to be queried.
     *
     * @return The height of the image in pixels.
     *
     * @throws IOException If an error occurs reading the height information
     *                     from the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public int getHeight(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	_checkImageIndex(imageIndex);

	// Return the image height
	NrrdMetadata metadata = new NrrdMetadata( _getHeader() );
	return metadata.getImageHeight();
    }

    /**
     * Returns an Iterator containing possible image types to which the given
     * image may be decoded, in the form of ImageTypeSpecifiers.  At least one
     * legal image type will be returned.
     *
     * @param imageIndex The index of the image to be retrieved.
     *
     * @return An Iterator containing at least one ImageTypeSpecifier
     *         representing suggested image types for decoding the current given
     *         image.
     *
     * @throws IOException If an error occurs reading the format information
     *                     from the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public Iterator getImageTypes(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	_checkImageIndex(imageIndex);

	// Get the image parameters
	NrrdMetadata metadata = new NrrdMetadata( _getHeader() );
	int width = getWidth(imageIndex);
	int height = getHeight(imageIndex);
	int bitsPerPixel = metadata.getBitsPerPixel();
	int type = metadata.getDataType();

	// Color images
	ArrayList list = new ArrayList(1);
	if ( metadata.hasColorImages() ) {
	    list.add( Utilities.getRgbImageType(width, height) );
	    return list.iterator();
	}

	// Determine the appropriate Image Type Specifier
	if (type == DataBuffer.TYPE_BYTE) {
	    list.add( Utilities.getGrayImageType(width, height, bitsPerPixel) );
	}
	else if (type == DataBuffer.TYPE_SHORT) {
	    list.add( Utilities.getGrayImageType(DataBuffer.TYPE_SHORT) );
	}
	else if (type == DataBuffer.TYPE_USHORT) {
	    list.add( Utilities.getGrayImageType(width, height, bitsPerPixel) );
	}
	else if (type == DataBuffer.TYPE_INT) {
	    list.add( Utilities.getGrayImageType(DataBuffer.TYPE_INT) );
	}
	else if (type == DataBuffer.TYPE_FLOAT) {
	    list.add( Utilities.getGrayImageType(DataBuffer.TYPE_FLOAT) );
	}
	else if (type == DataBuffer.TYPE_DOUBLE) {
	    list.add( Utilities.getGrayImageType(DataBuffer.TYPE_DOUBLE) );
	}

	// Check if the image data type is supported
	if (list.size() == 0) {
	    String msg = "Unsupported NRRD image type:  data type = " + type +
		"; bits per voxel = " + bitsPerPixel + ".";
	    throw new IOException(msg);
	}

	// Return the Image Type Specifier
	return list.iterator();
    }

    /**
     * Returns an IIOMetadata object representing the metadata associated with
     * the input source as a whole (i.e., not associated with any particular
     * image), or null if the reader does not support reading metadata, is set
     * to ignore metadata, or if no metadata is available.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     */
    public IIOMetadata getStreamMetadata() throws IOException
    {
	return new NrrdMetadata( _getHeader() );
    }

    /**
     * Returns an IIOMetadata object containing metadata associated with the
     * given image, or null if the reader does not support reading metadata, is
     * set to ignore metadata, or if no metadata is available.
     *
     * @param imageIndex Index of the image whose metadata is to be retrieved. 
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public IIOMetadata getImageMetadata(int imageIndex) throws IOException
    {
	// Not applicable
	return null;
    }

    /**
     * Reads the image indexed by imageIndex and returns it as a complete
     * BufferedImage.  The supplied ImageReadParam is ignored.
     *
     * @param imageIndex The index of the image to be retrieved.
     * @param param An ImageReadParam used to control the reading process, or
     *              null.
     *
     * @return The desired portion of the image as a BufferedImage.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public BufferedImage read(int imageIndex, ImageReadParam param)
	throws IOException
    {
	int width = getWidth(imageIndex);
	int height = getHeight(imageIndex);

	// Update the Listeners
	processImageStarted(imageIndex);

	// Get the number of bytes per pixel and byte swap
	NrrdMetadata metadata = new NrrdMetadata( _getHeader() );
	int bytesPerPixel = metadata.getBitsPerPixel()/8;
	boolean isByteSwapped = metadata.isByteSwapped();

	// Compute the size of the image in bytes
	ImageTypeSpecifier spec = (ImageTypeSpecifier)getImageTypes(0).next();
	int dataType = spec.getSampleModel().getDataType();
	int imageSize = width*height*bytesPerPixel;

	// Set the stream variables
	long imagePos = imageIndex*imageSize;
	int span = 1;

	// Adjust for non-domain first sample ordering
	if ( !metadata.varyDomainFirst() && !metadata.hasColorImages() ) {
	    int[] sizes = metadata.getSizes();

	    // Get the last dimension size
	    int lastSize = sizes[sizes.length-1];

	    // Divide the stream into blocks of size = last dimension size
	    int blockNumber = imageIndex % lastSize;

	    // Determine the offset within the block
	    int blockOffset = imageIndex / lastSize;

	    // Set the stream variables
	    imagePos = blockNumber*sizes[0]*imageSize +
		blockOffset*bytesPerPixel;
	    span = sizes[0];
	}

	// Return the raw image data if required
	if ( Utilities.returnRawBytes(param) ) {
	    BufferedImage bufferedImage = Utilities.getRawByteImage(imageSize);
	    DataBuffer db = bufferedImage.getRaster().getDataBuffer();
	    byte[] data = ((DataBufferByte)db).getData();
	    _readPixelData(imagePos, span, bytesPerPixel, data);
	    return bufferedImage;
	}

	// Create a Buffered Image for the image data
	Iterator iter = getImageTypes(0);
	BufferedImage bufferedImage = getDestination(null, iter, width, height);
	DataBuffer dataBuffer = bufferedImage.getRaster().getDataBuffer();

	// Read the image data into an array
	byte[] srcBuffer = new byte[imageSize];
	_readPixelData(imagePos, span, bytesPerPixel, srcBuffer);

	// RGB color (3-byte) image pixel data
	WritableRaster dstRaster = bufferedImage.getRaster();
	if ( metadata.hasColorImages() ) {
	    bufferedImage = Utilities.getRgbImage(srcBuffer, dstRaster, true);
	}

	// 8-bit gray image pixel data
	else if (dataBuffer.getDataType() == DataBuffer.TYPE_BYTE) {
	    bufferedImage = Utilities.getGrayByteImage(srcBuffer, dstRaster);
	}

	// 16-bit gray image pixel data (byte swap if necessary)
	else if (dataBuffer.getDataType() == DataBuffer.TYPE_USHORT) {
	    bufferedImage =
		Utilities.getGrayUshortImage(srcBuffer, dstRaster, 16, false,
					     isByteSwapped, 1, 0);
	}

	// General gray image pixel data (byte swap if necessary)
	else {
	    bufferedImage = Utilities.getGrayImage(srcBuffer, dstRaster,
						   isByteSwapped);
	}

	// Update the Listeners
	if ( abortRequested() ) { processReadAborted(); }
	else { processImageComplete(); }

	// Return the requested image
	return bufferedImage;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this ImageReader.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the input stream
	try {
	    if (input instanceof NrrdInputStream) {
		((NrrdInputStream)input).close();
	    }
	    else if (input instanceof ImageInputStream) {
		((ImageInputStream)input).close();
	    }
	}

	catch (Exception e) {}

	input = null;
	_pixelDataStream = null;
    }

    /**
     * Reads the NRRD Header from the input source.  If the NRRD Header has
     * already been decoded, no additional read is performed.
     *
     * @return NRRD Header decoded from the input source.
     * 
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     */
    private NrrdHeader _getHeader() throws IOException
    {
	// Decode the header if needed
	if (_header == null) {

	    // No input has been set
	    if (input == null) {
		String msg = "No input has been set.";
		throw new IllegalStateException(msg);
	    }

	    // Get the metadata Image Input Stream
	    ImageInputStream imageStream;
	    if (input instanceof NrrdInputStream) {
		NrrdInputStream inputStream = (NrrdInputStream)input;
		imageStream = inputStream.getMetadataInputStream();
	    }
	    else { imageStream = (ImageInputStream)input; }

	    // Decode the header
	    NrrdHeaderDecoder decoder = new NrrdHeaderDecoder(imageStream);
	    _header = decoder.getHeader();
	}

	return _header;
    }

    /**
     * Checks whether or not the specified image index is valid.
     *
     * @param imageIndex The index of the image to check.
     *
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    private void _checkImageIndex(int imageIndex) throws IOException
    {
	// Check the bounds of the image index
	if ( imageIndex >= getNumImages(false) ) {
	    String msg = "Image index of " + imageIndex + " >= the total " +
		"number of images (" + getNumImages(false) + ")";
	    throw new IndexOutOfBoundsException(msg);
	}
    }

    /**
     * Reads bytes of pixel data.
     *
     * @param pos Position in the pixel data stream to start reading at.
     * @param span Number of pixels separating adjacent pixel data.
     * @param bytesPerPixel Number of bytes per pixel data.
     * @param b Byte array to fill with pixel data.
     *
     * @throws IOException If an error occurs while reading from the pixel data
     *                     stream.
     * @throws IllegalStateException If the input has not been set.
     */
    private void _readPixelData(long pos, int span, int bytesPerPixel, byte[] b)
	throws IOException
    {
	// No input has been set
	if (input == null) {
	    String msg = "No input has been set.";
	    throw new IllegalStateException(msg);
	}

	// Pixel data input stream has not been set
	if (_pixelDataStream == null) {

	    // Get the Image Input Stream for image pixel data
	    ImageInputStream imageStream;
	    if (input instanceof NrrdInputStream) {
		NrrdInputStream inputStream = (NrrdInputStream)input;
		imageStream = inputStream.getPixelDataInputStream();
	    }
	    else { imageStream = (ImageInputStream)input; }

	    // Set the pixel data input stream
	    _pixelDataStream = new GZipImageInputStream(imageStream);
	}

	// Seek to the start
	_pixelDataStream.seek(pos);

	// Pixel data is in a block
	if (span == 1) {
	    _pixelDataStream.readFully(b);
	    return;
	}

	// Read the pixel data across blocks
	for (int i = 0; i < b.length/bytesPerPixel; i++) {

	    // Skip to the next pixel bytes
	    _pixelDataStream.seek(pos + i*span*bytesPerPixel);

	    // Read the pixel bytes
	    byte[] pb = new byte[bytesPerPixel];
	    _pixelDataStream.readFully(pb);

	    // Copy the bytes to the byte array
	    System.arraycopy(pb, 0, b, bytesPerPixel*i, bytesPerPixel);
	}
    }
}
