/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import edu.ucla.loni.imageio.AppletFriendlyIIOMetadata;
import edu.ucla.loni.imageio.BasicMetadataFormat;
import edu.ucla.loni.imageio.BasicMetadataTreeMaker;
import edu.ucla.loni.imageio.Utilities;
import edu.ucla.loni.nrrd.NrrdHeader;
import edu.ucla.loni.nrrd.NrrdHeaderElement;
import java.awt.image.DataBuffer;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.imageio.metadata.IIOInvalidTreeException;
import javax.imageio.metadata.IIOMetadataFormatImpl;
import javax.imageio.metadata.IIOMetadataNode;
import org.w3c.dom.Node;

/**
 * IIO Metadata for metadata of the NRRD image format.
 *
 * @version 15 March 2012
 */
public class NrrdMetadata extends AppletFriendlyIIOMetadata
{
    /** NRRD Header containing the original metadata. */
    private NrrdHeader _originalHeader;

    /** NRRD Header containing the current metadata. */
    private NrrdHeader _currentHeader;

    /**
     * Constructs an NRRD Metadata from the given NRRD Header.
     *
     * @param header NRRD Header containing the metadata.
     */
    public NrrdMetadata(NrrdHeader header)
    {
	super(true, NrrdMetadataFormat.NATIVE_METADATA_FORMAT_NAME,
	      NrrdMetadataFormat.class.getName(),
	      new String[]{BasicMetadataFormat.NATIVE_METADATA_FORMAT_NAME},
	      new String[]{BasicMetadataFormat.class.getName()});

	_originalHeader = header;
	_currentHeader = header;
    }

    /**
     * Returns true if this object does not support the mergeTree, setFromTree,
     * and reset methods.  
     *
     * @return True if this IIOMetadata object cannot be modified; false o/w.
     */
    public boolean isReadOnly()
    {
	return false;
    }

    /**
     * Returns an XML DOM Node object that represents the root of a tree of
     * metadata contained within this object according to the conventions
     * defined by a given metadata format.
     *
     * @param formatName Name of the desired metadata format.
     *
     * @return An XML DOM Node object forming the root of a tree.
     *
     * @throws IllegalArgumentException If formatName is not one of the allowed
     *                                  metadata format names.
     */
    public Node getAsTree(String formatName)
    {
	// Standard format name
	String standardName = IIOMetadataFormatImpl.standardMetadataFormatName;
	if ( standardName.equals(formatName) ) { return getStandardTree(); }

	// Native metadata format name
	if ( nativeMetadataFormatName.equals(formatName) ) {
	    return NrrdMetadataConversions.toTree(_currentHeader);
	}

	// LONI Basic format name
	String basicName = BasicMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( basicName.equals(formatName) ) { return _getBasicTree(); }

	// Unknown format name
	String msg = "The format name \"" + formatName + "\" is not a " +
	    "supported metadata format name.";
	throw new IllegalArgumentException(msg);
    }

    /**
     * Alters the internal state of this IIOMetadata object from a tree of XML
     * DOM Nodes whose syntax is defined by the given metadata format.  The
     * previous state is altered only as necessary to accomodate the nodes that
     * are present in the given tree.
     *
     * @param formatName Name of the desired metadata format.
     * @param root An XML DOM Node object forming the root of a tree.
     *
     * @throws IllegalArgumentException If formatName is not one of the allowed
     *                                  metadata format names, or if the root is
     *                                  null.
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     */
    public void mergeTree(String formatName, Node root)
	throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check for a matching root Node
	if ( !root.getNodeName().equals(nativeMetadataFormatName) ) {
	    String msg = "Root node must be named \"" +
		nativeMetadataFormatName + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// Convert the root
	NrrdHeader header = NrrdMetadataConversions.toNrrdHeader(root);

	// Replace the current NRRD Header if none exists
	if (_currentHeader == null) {
	    _currentHeader = header;
	    return;
	}

	// Otherwise merge headers
	NrrdHeaderElement[] elements = header.getElements();
	for (int i = 0; i < elements.length; i++) {
	    _currentHeader.addElement(elements[i]);
	}
    }

    /**
     * Resets all the data stored in this object to default values, usually to
     * the state this object was in immediately after construction, though the
     * precise semantics are plug-in specific.  Note that there are many
     * possible default values, depending on how the object was created.
     */
    public void reset()
    {
	// Reset the current Header to the original 
	_currentHeader = _originalHeader;
    }

    /**
     * Gets the number of images.
     *
     * @return Number of images.
     *
     * @throws IllegalStateException If there are more than 4 dimensions.
     */
    public int getNumberOfImages()
    {
	int[] sizes = getSizes();

	// Only support 4 dimensions
	if (sizes.length > 4) {
	    throw new IllegalStateException("Too many dimensions (>4)");
	}

	// Return the number of images
	if (sizes.length == 2) { return 1; }
	if (sizes.length == 3) {
	    if ( hasColorImages() ) { return 1; }
	    return sizes[2];
	}
	if ( varyDomainFirst() ) { return sizes[2]*sizes[3]; }
	if ( hasColorImages() ) { return sizes[3]; }
	return sizes[0]*sizes[3];
    }

    /**
     * Gets the width of each image.
     *
     * @return Width of each image.
     */
    public int getImageWidth()
    {
	int[] sizes = getSizes();

	// Return the image width
	if (sizes.length <= 3) {
	    if ( hasColorImages() ) { return sizes[1]; }
	    return sizes[0];
	}
	if ( varyDomainFirst() ) { return sizes[0]; }
	return sizes[1];
    }

    /**
     * Gets the height of each image.
     *
     * @return Height of each image.
     */
    public int getImageHeight()
    {
	int[] sizes = getSizes();

	// Return the image height
	if (sizes.length <= 3) {
	    if ( hasColorImages() ) { return sizes[2]; }
	    return sizes[1];
	}
	if ( varyDomainFirst() ) { return sizes[1]; }
	return sizes[2];
    }

    /**
     * Determines if the pixel data is byte-swapped.
     *
     * @return True if the pixel data is byte-swapped; false otherwise.
     */
    public boolean isByteSwapped()
    {
	String value = _currentHeader.getFieldDescriptor("endian");
	return "little".equals(value);
    }

    /**
     * Determines if the pixel data is gzip compressed.
     *
     * @return True if the pixel data is gzip compressed; false otherwise.
     */
    public boolean isGzipCompressed()
    {
	String value = _currentHeader.getFieldDescriptor("encoding");
	return "gz".equals(value) || "gzip".equals(value);
    }

    /**
     * Gets the number of bits per pixel used by each image.
     *
     * @return Number of bits per pixel used by each image, or -1 if it cannot
     *         be determined.
     */
    public int getBitsPerPixel()
    {
	String type = _getBasicDataType();
	if (type == null) { return -1; }

	// RGB (3 bytes/voxel)
	if ( hasColorImages() ) { return 24; }

	// Byte
	if ( type.equals(BasicMetadataFormat.BYTE_TYPE) ) { return 8; }

	// Short
	if ( type.equals(BasicMetadataFormat.SIGNED_SHORT_TYPE) ||
	     type.equals(BasicMetadataFormat.UNSIGNED_SHORT_TYPE) )
	    {
		return 16;
	    }

	// Integer
	if ( type.equals(BasicMetadataFormat.SIGNED_INT_TYPE) ||
	     type.equals(BasicMetadataFormat.UNSIGNED_INT_TYPE) )
	    {
		return 32;
	    }

	// Float
	if ( type.equals(BasicMetadataFormat.FLOAT_TYPE) ) { return 32; }

	// Double
	if ( type.equals(BasicMetadataFormat.DOUBLE_TYPE) ) { return 64; }

	// Unknown type
	return -1;
    }

    /**
     * Gets the data type of each image.
     *
     * @return Data type or -1 if it cannot be determined.
     */
    public int getDataType()
    {
	String type = _getBasicDataType();
	if (type == null) { return -1; }

	// Byte
	if ( type.equals(BasicMetadataFormat.BYTE_TYPE) ) {
	    return DataBuffer.TYPE_BYTE;
	}

	// Short
	if ( type.equals(BasicMetadataFormat.SIGNED_SHORT_TYPE) ) {
	    return DataBuffer.TYPE_SHORT;
	}

	// Unsigned short
	if ( type.equals(BasicMetadataFormat.UNSIGNED_SHORT_TYPE) ) {
	    return DataBuffer.TYPE_USHORT;
	}

	// Integer
	if ( type.equals(BasicMetadataFormat.SIGNED_INT_TYPE) ||
	     type.equals(BasicMetadataFormat.UNSIGNED_INT_TYPE) )
	    {
		return DataBuffer.TYPE_INT;
	    }

	// Float
	if ( type.equals(BasicMetadataFormat.FLOAT_TYPE) ) {
	    return DataBuffer.TYPE_FLOAT;
	}

	// Double
	if ( type.equals(BasicMetadataFormat.DOUBLE_TYPE) ) {
	    return DataBuffer.TYPE_DOUBLE;
	}

	// Unknown type
	return -1;
    }

    /**
     * Determines whether or not each image is a color image.
     *
     * @return True if the images are color images; false otherwise.
     *
     * @throws IllegalArgumentException If the color type is unsupported.
     */
    public boolean hasColorImages()
    {
	// Get the dimension types
	String value = _currentHeader.getFieldDescriptor("kinds");
	StringTokenizer tokenizer = new StringTokenizer(value, " ");

	// Search for a color type
	boolean isFirstType = true;
	while ( tokenizer.hasMoreTokens() ) {
	    String type = tokenizer.nextToken();

	    // Supported RGB
	    if (isFirstType) {
		if ("3-color".equals(type)) { return true; }
		if ("RGB-color".equals(type)) { return true; }
	    }

	    // Unsupported RGB
	    if (type.indexOf("color") != -1) {
		String msg = "Unsupported color type: " + value;
		throw new IllegalArgumentException(msg);
	    }

	    isFirstType = false;
	}

	return false;
    }

    /**
     * Gets the size of each dimension.
     *
     * @return Size of each dimension.
     */
    public int[] getSizes()
    {
	try {

	    // Parse the size string
	    String value = _currentHeader.getFieldDescriptor("sizes");
	    StringTokenizer tokenizer = new StringTokenizer(value, " ");
	    
	    // Get the string sizes
	    Vector stringSizes = new Vector();
	    while ( tokenizer.hasMoreTokens() ) {
		stringSizes.add( tokenizer.nextToken() );
	    }

	    // Convert to integers
	    int[] sizes = new int[stringSizes.size()];
	    for (int i = 0; i < sizes.length; i++) {
		sizes[i] = Integer.parseInt((String)stringSizes.get(i));
	    }


	    // Return the sizes
	    return sizes;
	}

	catch (Exception e) { return new int[0]; }
    }

    /**
     * Determines if the voxels vary first by domain.
     *
     * @return True if the voxels vary first by domain; false if the voxels vary
     *         first by a non-domain dimension.
     *
     * @throws IllegalArgumentException If the voxels vary by neither option.
     */
    public boolean varyDomainFirst()
    {
	// Get the dimension types
	String value = _currentHeader.getFieldDescriptor("kinds");
	StringTokenizer tokenizer = new StringTokenizer(value, " ");

	// Get the types
	Vector types = new Vector();
	while ( tokenizer.hasMoreTokens() ) {
	    types.add( tokenizer.nextToken() );
	}

	// First type is not a domain
	if ( !_isDomain((String)types.get(0)) ) {

	    // Other types must be domains
	    for (int i = 1; i < types.size(); i++) {
		if ( !_isDomain((String)types.get(i)) ) {
		    throw new IllegalArgumentException("Unsupported: " + value);
		}
	    }

	    // Non-domain
	    return false;
	}

	// First types are domains
	for (int i = 0; i < types.size()-1; i++) {
	    if ( !_isDomain((String)types.get(i)) ) {
		throw new IllegalArgumentException("Unsupported: " + value);
	    }
	}

	// Domain
	return true;
    }

    /**
     * Returns an IIOMetadataNode representing the chroma information of the
     * standard javax_imageio_1.0 metadata format, or null if no such
     * information is available.
     *
     * @return IIOMetadataNode representing the chroma information, or null if
     *         it does not exist.
     */
    protected IIOMetadataNode getStandardChromaNode()
    {
	try {

	    // RGB
	    if ( hasColorImages() ) { return Utilities.getRgbChromaNode(); }

	    // Only support non-inverted grayscale
	    return Utilities.getGrayScaleChromaNode(false);
	}

	// No information available
	catch (Exception e) { return null; }
    }

    /**
     * Returns an IIOMetadataNode representing the data format information of
     * the standard javax_imageio_1.0 metadata format, or null if no such
     * information is available.
     *
     * @return IIOMetadataNode representing the data information, or null if
     *         it does not exist.
     */
    protected IIOMetadataNode getStandardDataNode()
    {
	try {
	    int type = getDataType();

	    // RGB
	    if ( hasColorImages() ) { return Utilities.getRgbDataNode(); }

	    // Byte
	    if (type == DataBuffer.TYPE_BYTE) {
		return Utilities.getGrayScaleDataNode(8);
	    }

	    // Short
	    if (type == DataBuffer.TYPE_SHORT ||
		type == DataBuffer.TYPE_USHORT)
		{
		    return Utilities.getGrayScaleDataNode(16);
		}

	    // Integer
	    if (type == DataBuffer.TYPE_INT) {
		return Utilities.getGrayScaleDataNode(32);
	    }

	    // Float
	    if (type == DataBuffer.TYPE_FLOAT) {
		return Utilities.getGrayScaleDataNode(32, false);
	    }

	    // Double
	    if (type == DataBuffer.TYPE_DOUBLE) {
		return Utilities.getGrayScaleDataNode(64, false);
	    }

	    // Unknown
	    return null;
	}
      
	// No information available
	catch (Exception e) { return null; }
    }

    /**
     * Returns an IIOMetadataNode representing the dimension format information
     * of the standard javax_imageio_1.0 metadata format, or null if no such
     * information is available.
     *
     * @return IIOMetadataNode representing the dimension information, or null
     *         if it does not exist.
     */
    protected IIOMetadataNode getStandardDimensionNode()
    {
	try {

	    // Get the pixel dimensions
	    float[] pixelDims = _getPixelDimensions();

	    // Return the dimension node
	    return Utilities.getDimensionNode(pixelDims[0], pixelDims[1]);
	}

	// Unable to make the determination
	catch (Exception e) { return null; }
    }

    /**
     * Gets the LONI Basic metadata tree.
     *
     * @return DOM Node representing the LONI Basic metadata, or null if it can
     *         not be determined.
     */
    private Node _getBasicTree()
    {
	// Image dimensions
	int imageWidth = getImageWidth();
	int imageHeight = getImageHeight();
	int numberOfImages = getNumberOfImages();

	// Pixel dimensions
	float[] pixelDims = _getPixelDimensions();
	float pixelWidth = pixelDims[0];
	float pixelHeight = pixelDims[1];
	float sliceThickness = 0;
	if (pixelDims.length > 2) { sliceThickness = pixelDims[2]; }

	// Data type
	String dataType = _getBasicDataType();
	if (dataType == null) { dataType = "Unknown"; }

	// Bits per pixel
	int bitsPerPixel = getBitsPerPixel();

	// Color type
	String colorType = BasicMetadataFormat.GRAY_TYPE;
	if ( hasColorImages() ) { colorType = BasicMetadataFormat.RGB_TYPE; }

	// Return the basic tree
	return BasicMetadataTreeMaker.getTree(imageWidth, imageHeight,
					      numberOfImages, pixelWidth,
					      pixelHeight, sliceThickness,
					      dataType, bitsPerPixel,
					      colorType);
    }

    /**
     * Gets the basic data type.
     *
     * @return Basic data type or null if it cannot be determined.
     */
    private String _getBasicDataType()
    {
	String type = _currentHeader.getFieldDescriptor("type");

	// Signed 1-byte
	if ("signed char".equals(type) || "int8".equals(type) ||
	    "int8_t".equals(type))
	    {
		return BasicMetadataFormat.BYTE_TYPE;
	    }

	// Unsigned 1-byte
	if ("uchar".equals(type) || "unsigned char".equals(type) ||
	    "uint8".equals(type) || "uint8_t".equals(type))
	    {
		return BasicMetadataFormat.BYTE_TYPE;
	    }

	// Signed 2-bytes (short)
	if ("short".equals(type) || "short int".equals(type) ||
	    "signed short".equals(type) ||
	    "signed short int".equals(type) || "int16".equals(type) ||
	    "int16_t".equals(type))
	    {
		return BasicMetadataFormat.SIGNED_SHORT_TYPE;
	    }

	// Unsigned 2-bytes (short)
	if ("ushort".equals(type) || "unsigned short".equals(type) ||
	    "unsigned short int".equals(type) || "uint16".equals(type) ||
	    "uint16_t".equals(type))
	    {
		return BasicMetadataFormat.UNSIGNED_SHORT_TYPE;
	    }

	// Signed 4-bytes (integer)
	if ("int".equals(type) || "signed int".equals(type) ||
	    "int32".equals(type) || "int32_t".equals(type))
	    {
		return BasicMetadataFormat.SIGNED_INT_TYPE;
	    }

	// Unsigned 4-bytes (integer)
	if ("uint".equals(type) || "unsigned int".equals(type) ||
	    "uint32".equals(type) || "uint32_t".equals(type))
	    {
		return BasicMetadataFormat.UNSIGNED_INT_TYPE;
	    }

	// Float
	if ("float".equals(type)) { return BasicMetadataFormat.FLOAT_TYPE; }

	// Double
	if ("double".equals(type)) { return BasicMetadataFormat.DOUBLE_TYPE; }

	// Unknown type
	return null;
    }

    /**
     * Gets the pixel dimensions.
     *
     * @return Pixel dimensions.
     */
    private float[] _getPixelDimensions()
    {
	// Get the dimension vectors
	String value = _currentHeader.getFieldDescriptor("space directions");
	StringTokenizer tokenizer = new StringTokenizer(value, " ");

	// Get the vectors
	Vector vectors = new Vector();
	while ( tokenizer.hasMoreTokens() ) {
	    String token = tokenizer.nextToken();
	    if ( !"none".equals(token) ) { vectors.add(token); }
	}

	// Get the first vector's components
	String vector1 = (String)vectors.get(0);
	tokenizer = new StringTokenizer(vector1, "(,)");
	float x1 = Float.parseFloat( tokenizer.nextToken() );
	float y1 = Float.parseFloat( tokenizer.nextToken() );
	float z1 = Float.parseFloat( tokenizer.nextToken() );

	// Magnitude of the first vector is the voxel width
	float pixelWidth = (float)Math.sqrt(x1*x1 + y1*y1 + z1*z1);

	// Get the second vector's components
	String vector2 = (String)vectors.get(1);
	tokenizer = new StringTokenizer(vector2, "(,)");
	float x2 = Float.parseFloat( tokenizer.nextToken() );
	float y2 = Float.parseFloat( tokenizer.nextToken() );
	float z2 = Float.parseFloat( tokenizer.nextToken() );

	// Magnitude of the second vector is the voxel height
	float pixelHeight = (float)Math.sqrt(x2*x2 + y2*y2 + z2*z2);

	// Only 2 dimensions
	if (vectors.size() == 2) {
	    return new float[]{pixelWidth, pixelHeight};
	}

	// Get the third vector's components
	String vector3 = (String)vectors.get(2);
	tokenizer = new StringTokenizer(vector3, "(,)");
	float x3 = Float.parseFloat( tokenizer.nextToken() );
	float y3 = Float.parseFloat( tokenizer.nextToken() );
	float z3 = Float.parseFloat( tokenizer.nextToken() );

	// Magnitude of the third vector is the slice spacing
	float pixelLength = (float)Math.sqrt(x3*x3 + y3*y3 + z3*z3);

	// Return the pixel dimensions
	return new float[]{pixelWidth, pixelHeight, pixelLength};
    }

    /**
     * Determines if the dimension is an independent variable.
     *
     * @param kind Dimension label.
     *
     * @return True if the dimension is an independent variable; false o/w.
     */
    private static boolean _isDomain(String kind)
    {
	if ( "domain".equals(kind) ) { return true; }
	if ( "space".equals(kind) ) { return true; }
	if ( "time".equals(kind) ) { return true; }

	return false;
    }
}
