/*
  COPYRIGHT NOTICE
  Copyright (c) 20010  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import java.io.IOException;
import java.io.OutputStream;
import javax.imageio.stream.ImageOutputStream;

/**
 * Class that adapts the Image Output Stream interface into the Output Stream
 * class.
 *
 * @version 29 October 2010
 */
public class OutputStreamAdapter extends OutputStream
{
    /** Image Output Stream to adapt. */
    private ImageOutputStream _imageOutputStream;

    /**
     * Creates an Output Stream Adapter.
     *
     * @param imageOutputStream Image Output Stream to adapt.
     *
     * @throws NullPointerException If the Image Output Stream is null.
     */
    public OutputStreamAdapter(ImageOutputStream imageOutputStream)
    {
	// Check for a null stream
	if (imageOutputStream == null) {
	    String msg = "A null Image Output Stream is not allowed.";
	    throw new NullPointerException(msg);
	}

	_imageOutputStream = imageOutputStream;
    }

    /**
     * Writes the specified byte to this output stream.
     *
     * @param b Byte to write.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void write(int b) throws IOException
    {
	_imageOutputStream.write(b);
    }

    /**
     * Writes len bytes from the specified byte array starting at the offset to
     * this output stream.
     *
     * @param b Buffer of data to write.
     * @param off The start offset in array b at which the data is read.
     * @param len The number of bytes to write.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void write(byte b[], int off, int len) throws IOException
    {
	_imageOutputStream.write(b, off, len);
    }

    /**
     * Flushes this output stream and forces any buffered output bytes to be
     * written out.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void flush() throws IOException
    {
	_imageOutputStream.flush();
    }

    /**
     * Closes this output stream and releases any system resources associated
     * with the stream.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void close() throws IOException
    {
	_imageOutputStream.close();
    }
}
